"""Constants used in Django Channels."""

# Channel used for notifying storage manager to start.
CHANNEL_STORAGE_MANAGER_WORKER = "storage.manager"
# Channel used for notifying cleanup manager to start.
CHANNEL_STORAGE_CLEANUP_WORKER = "storage.cleanup"
# Message type for starting the storage manager.
TYPE_STORAGE_MANAGER_RUN = "storagemanager.run"
# Message type for starting the storage cleanup.
TYPE_STORAGE_CLEANUP_RUN = "storagecleanup.run"



"""Resolwe collection model."""

from django.contrib.contenttypes.models import ContentType
from django.contrib.postgres.aggregates import ArrayAgg
from django.contrib.postgres.fields import ArrayField
from django.contrib.postgres.indexes import GinIndex
from django.contrib.postgres.search import SearchVectorField
from django.db import models

from resolwe.flow.models.base import BaseQuerySet
from resolwe.observers.consumers import BackgroundTaskType
from resolwe.observers.models import BackgroundTask
from resolwe.observers.utils import start_background_task
from resolwe.permissions.models import PermissionObject

from .annotations import AnnotationField
from .base import BaseModel, DataStatus
from .history_manager import HistoryMixin
from .utils import DirtyError, validate_schema


class BaseCollection(BaseModel):
    """Template for Postgres model for storing a collection."""

    class Meta(BaseModel.Meta):
        """BaseCollection Meta options."""

        abstract = True

    #: detailed description
    description = models.TextField(blank=True)

    settings = models.JSONField(default=dict)

    #: tags for categorizing objects
    tags = ArrayField(models.CharField(max_length=255), default=list)

    #: field used for full-text search
    search = SearchVectorField(null=True)


class BaseCollectionQuerySet(BaseQuerySet):
    """A base queryset used by both collection and entity."""

    def annotate_status(self):
        """Annotate the queryset with status of the collections."""

        priority_order = models.Case(
            *[
                models.When(data__status=status, then=models.Value(sort_order))
                for sort_order, status in enumerate(DataStatus.sort_order())
            ]
        )
        return self.annotate(
            statuses=ArrayAgg(models.F("data__status"), ordering=priority_order.asc())
        ).annotate(status=models.F("statuses__0"))


class CollectionQuerySet(BaseCollectionQuerySet):
    """Query set for ``Collection`` objects."""

    def duplicate(self, request_user) -> BackgroundTask:
        """Duplicate (make a copy) ``Collection`` objects in the background."""
        task_data = {"collection_ids": list(self.values_list("pk", flat=True))}
        return start_background_task(
            BackgroundTaskType.DUPLICATE_COLLECTION,
            "Duplicate collections",
            task_data,
            request_user,
        )

    def delete_background(self, request_user):
        """Delete the ``Collection`` objects in the background."""
        task_data = {
            "object_ids": list(self.values_list("pk", flat=True)),
            "content_type_id": ContentType.objects.get_for_model(self.model).pk,
        }
        return start_background_task(
            BackgroundTaskType.DELETE, "Delete collections", task_data, request_user
        )


class Collection(HistoryMixin, PermissionObject, BaseCollection):
    """Postgres model for storing a collection."""

    class Meta(BaseCollection.Meta):
        """Collection Meta options."""

        permissions = (
            ("view", "Can view collection"),
            ("edit", "Can edit collection"),
            ("share", "Can share collection"),
            ("owner", "Is owner of the collection"),
        )

        indexes = [
            models.Index(name="idx_collection_name", fields=["name"]),
            GinIndex(
                name="idx_collection_name_trgm",
                fields=["name"],
                opclasses=["gin_trgm_ops"],
            ),
            models.Index(name="idx_collection_slug", fields=["slug"]),
            GinIndex(name="idx_collection_tags", fields=["tags"]),
            GinIndex(name="idx_collection_search", fields=["search"]),
        ]

    #: manager
    objects = CollectionQuerySet.as_manager()

    #: duplication date and time
    duplicated = models.DateTimeField(blank=True, null=True)

    #: annotation fields available to samples in this collection
    annotation_fields = models.ManyToManyField(
        AnnotationField, related_name="collection"
    )

    #: collection descriptor schema
    descriptor_schema = models.ForeignKey(
        "flow.DescriptorSchema", blank=True, null=True, on_delete=models.PROTECT
    )

    #: collection descriptor
    descriptor = models.JSONField(default=dict)

    #: indicate whether `descriptor` doesn't match `descriptor_schema` (is dirty)
    descriptor_dirty = models.BooleanField(default=False)

    def is_duplicate(self):
        """Return True if collection is a duplicate."""
        return bool(self.duplicated)

    def duplicate(self, request_user) -> BackgroundTask:
        """Duplicate (make a copy) object in the background."""
        task_data = {"collection_ids": [self.pk]}
        return start_background_task(
            BackgroundTaskType.DUPLICATE_COLLECTION,
            "Duplicate collection",
            task_data,
            request_user,
        )

    def delete_background(self):
        """Delete the object in the background."""
        task_data = {
            "object_ids": [self.pk],
            "content_type_id": ContentType.objects.get_for_model(self).pk,
        }
        return start_background_task(
            BackgroundTaskType.DELETE, "Delete collections", task_data, self.contributor
        )

    def save(self, *args, **kwargs):
        """Perform descriptor validation and save object."""
        create = self.pk is None
        if self.descriptor_schema:
            try:
                validate_schema(self.descriptor, self.descriptor_schema.schema)
                self.descriptor_dirty = False
            except DirtyError:
                self.descriptor_dirty = True
        elif self.descriptor and self.descriptor != {}:
            raise ValueError(
                "`descriptor_schema` must be defined if `descriptor` is given"
            )
        super().save(*args, **kwargs)
        if create:
            required_fields = AnnotationField.objects.filter(required=True)
            if required_fields.exists():
                self.annotation_fields.add(*required_fields)



"""History manager.

TODO: move it to a 1different namespace.
"""

from collections import defaultdict
from datetime import datetime
from typing import TYPE_CHECKING, Any, Optional, Type

from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.db.backends.postgresql.psycopg_any import DateTimeTZRange
from django.utils import timezone

if TYPE_CHECKING:
    from .history import TrackChange

    _Base = models.Model
else:
    _Base = object


class HistoryManager:
    """Keeps track of the history changes.

    All tracked fields are registered automatically and changes must be tracked through
    this model.
    """

    def __init__(self):
        """Initialize mapping between field name and tracker."""
        self._trackers: dict[tuple[Type[models.Model], str], Type["TrackChange"]] = (
            dict()
        )
        self._tracked_fields: dict[Type[models.Model], list[str]] = defaultdict(list)

    def add_tracker(self, tracker: Type["TrackChange"]):
        """Add tracker to the manager.

        :raises AssertionError: when tracker for the same field already exists.
        """
        for model, name in tracker.applicable_to:
            assert (
                model,
                name,
            ) not in self._trackers, (
                f"Tracker for field '{name}' and model '{model}' already exists."
            )
            self._trackers[(model, name)] = tracker
            self._tracked_fields[model].append(name)

    def valid_value(
        self,
        instance: "HistoryMixin",
        field_name: str,
        timestamp: Optional[datetime] = None,
    ) -> Any:
        """Get the valid value at the point in time.

        :attr timestamp: a point in time. When not provided the current moment is used.
        :raises AttributeError: when no value exists for the given timestamp.
        """
        # Take the latest change before the timestamp. When timestamp is not provided
        # take the latest change.
        history = instance.get_history_object()
        tracker = self._trackers[(instance._meta.model, field_name)]
        if timestamp:
            tracker_qset = tracker.objects.filter(
                history=history, timestamp__lte=timestamp
            )
        else:
            tracker_qset = tracker.objects.filter(history=history)

        change = tracker_qset.last()
        if change is None:
            raise RuntimeError(
                f"No value exists for history object {history.pk}, field "
                f"{tracker.applicable_to} and timestamp {timestamp}."
            )
        return change.get_value(field_name)

    def process_change(self, instance: "HistoryMixin"):
        """Process the change in the field.

        :raises KeyError: when tracker for the field is not registered.
        :raises AssertionError: when the instance does not have history tracking.
        """
        model = instance._meta.model
        for field_name in self._tracked_fields[model]:
            if (model, field_name) not in self._trackers:
                continue
            try:
                tracker = self._trackers[(model, field_name)]
                if hasattr(instance, field_name):
                    current_value = self.valid_value(instance, field_name)
                    new_value = getattr(instance, field_name)
                    if current_value != new_value:
                        tracker.process_change(instance, field_name)
            except RuntimeError:
                tracker.process_change(instance, field_name)


history_manager = HistoryManager()


class HistoryMixin(_Base):
    """Extend this model to track its changes."""

    # For typing purposes.
    history: models.Manager

    def delete(self, *args, **kwargs):
        """Delete the object and update the history object."""
        history_object = self.get_history_object()
        super().delete(*args, **kwargs)
        history_object._meta.model.objects.filter(pk=history_object.pk).update(
            deleted=timezone.now()
        )

    def save(self, *args, **kwargs):
        """Save the object and update the history object."""
        super().save(*args, **kwargs)
        history_manager.process_change(self)

    def get_history_object(self):
        """Get the valid history object for the instance.

        If the object does not exist yet, create it.
        """
        try:
            return self.history.get(valid__upper_inf=True)
        except ObjectDoesNotExist:
            return self.history.create(valid=DateTimeTZRange(lower=timezone.now()))

    def get_history_model(self):
        """Get the model of the history object."""
        return self.history.model



"""Safe Resolwe process parser."""

import ast
import collections
import json
from collections import defaultdict
from contextlib import suppress
from inspect import isclass
from typing import Iterable, Mapping, Set

import asteval
from django.conf import settings

from . import runtime
from .descriptor import Persistence, ProcessDescriptor, SchedulingClass
from .fields import get_available_fields
from .runtimes import python_runtimes_manager


class StaticMetadata:
    """Process metadata item."""

    def get_value(self, node):
        """Convert value from an AST node."""
        raise NotImplementedError


class StaticStringMetadata(StaticMetadata):
    """String metadata item."""

    def get_value(self, node):
        """Convert value from an AST node."""
        if not isinstance(node, ast.Str):
            raise TypeError("must be a string literal")

        return node.s


class StaticEnumMetadata(StaticMetadata):
    """Enum metadata item."""

    def __init__(self, choices):
        """Construct enum metadata item."""
        self.choices = choices

    def get_value(self, node):
        """Convert value from an AST node."""
        if not isinstance(node, ast.Attribute):
            raise TypeError("must be an attribute")

        if node.value.id != self.choices.__name__:
            raise TypeError("must be an attribute of {}".format(self.choices.__name__))

        return getattr(self.choices, node.attr)


class StaticDictMetadata(StaticMetadata):
    """Dict metadata item."""

    def get_value(self, node):
        """Convert value from an AST node."""
        if not isinstance(node, ast.Dict):
            raise TypeError("must be a dictionary")

        evaluator = SafeEvaluator()
        try:
            value = evaluator.run(node)
        except Exception as ex:
            # TODO: Handle errors.
            raise ex

        try:
            # Ensure value is a serializable dictionary.
            value = json.loads(json.dumps(value))
            if not isinstance(value, dict):
                raise TypeError
        except (TypeError, ValueError) as exception:
            raise TypeError("must be serializable") from exception

        return value


# Possible process metadata.
PROCESS_METADATA = {
    "slug": StaticStringMetadata(),
    "name": StaticStringMetadata(),
    "process_type": StaticStringMetadata(),
    "version": StaticStringMetadata(),
    "category": StaticStringMetadata(),
    "scheduling_class": StaticEnumMetadata(choices=SchedulingClass),
    "persistence": StaticEnumMetadata(choices=Persistence),
    "requirements": StaticDictMetadata(),
    "data_name": StaticStringMetadata(),
    "entity": StaticDictMetadata(),
}


class SafeEvaluator(asteval.Interpreter):
    """Safe evaluator of Python expressions."""

    def __init__(self, symtable=None):
        """Initialize the safe evaluator."""
        if symtable is None:
            # Ensure symbol table is safely initialized as leaving it set to None makes
            # all kinds of builtin functions available.
            symtable = {}

        super().__init__(symtable=symtable, use_numpy=False, no_print=True)


class ProcessVisitor(ast.NodeVisitor):
    """Visitor for parsing process AST."""

    def __init__(self, source):
        """Construct process AST visitor."""
        self.source = source
        self.processes = []
        self.base_classes = set()
        # Mapping betwwen process slug and AST node representing the class.
        self.ast_nodes = dict()

        super().__init__()

    def _get_called_slug(self, process_slug, called_node: ast.Call):
        """Get the slug of the process being called.

        :raises RuntimeError: when slug could not be determined.
        """
        # Both methods (run_process and get_latest) take argument slug as the first
        # argument, so no separation is necessary.
        # When the first non-keyword argumet is given it must be slug.
        keyword_map = {keyword.arg: keyword.value for keyword in called_node.keywords}
        slug = (
            called_node.args[0]
            if len(called_node.args) >= 1
            else keyword_map.get("slug")
        )
        if isinstance(slug, (ast.Constant, ast.Str)):
            return getattr(slug, "value", slug.s), ast.Constant

        if isinstance(slug, ast.Name) and isinstance(slug.ctx, ast.Load):
            return slug.id, ast.Name

        raise RuntimeError(
            f"Spaned process slug could not be determined for the process {process_slug}."
        )

    def get_possible_variable_values(
        self, process_node: ast.ClassDef
    ) -> Mapping[str, Iterable[str]]:
        """Get the mapping between variable names and their possible values.

        The variable is only included in the mapping when it is assigned only
        constant values.
        """
        mapping = defaultdict(list)
        unknown_values: Set[str] = set()
        for child_node in ast.walk(process_node):
            if isinstance(child_node, ast.Assign):
                if isinstance(child_node.value, (ast.Constant, ast.Str)):
                    for target in child_node.targets:
                        if isinstance(target, ast.Name):
                            mapping[target.id].append(
                                getattr(child_node.value, "value", child_node.value.s)
                            )
                else:
                    unknown_values.update(
                        target.id
                        for target in child_node.targets
                        if isinstance(target, ast.Name)
                    )
        # Remove the variables with non-constant assignment from the mapping.
        for unknown in unknown_values:
            mapping.pop(unknown, None)
        return mapping

    def get_dependencies(self):
        """Get the dependencies between the processes.

        :raises RuntimeError: when dependencies could not be determined. For
            instance when process slug is a variable.

        :returns: the dictionary where key is the process slug and the
            corresponding value is the list of processes the process can spawn.
        """
        slugs = dict()
        # Register the runtime with the full name and only the class name.
        # For example "resolwe.process.runtime.Process" and "Process".
        candidate_names = ["self.run_process"]
        for candidate in settings.FLOW_PROCESSES_RUNTIMES:
            candidate_names.append(f"{candidate}.get_latest")
            candidate_names.append(f"{candidate}.get")
            candidate_names.append(f"{candidate.split('.')[-1]}.get_latest")
            candidate_names.append(f"{candidate.split('.')[-1]}.get")
        for process_slug, process_node in self.ast_nodes.items():
            mapping = self.get_possible_variable_values(process_node)
            slugs[process_slug] = set()
            for child_node in ast.walk(process_node):
                if isinstance(child_node, ast.Call):
                    with suppress(AttributeError):
                        candidate_name = (
                            f"{child_node.func.value.id}.{child_node.func.attr}"
                        )
                        if candidate_name in candidate_names:
                            called_slug, slug_type = self._get_called_slug(
                                process_slug, child_node
                            )
                            # The slug is a variable. Collect its possible names.
                            if slug_type == ast.Name:
                                if called_slug not in mapping:
                                    raise RuntimeError(
                                        f"Unable to determine value for the "
                                        f"variable '{called_slug}' in the "
                                        f"process '{process_slug}'."
                                    )
                                slugs[process_slug].update(mapping[called_slug])
                            if slug_type == ast.Constant:
                                slugs[process_slug].add(called_slug)
        return slugs

    def visit_field_class(self, item, descriptor=None, fields=None):
        """Visit a class node containing a list of field definitions."""
        discovered_fields = collections.OrderedDict()
        field_groups = {}
        for node in item.body:
            if isinstance(node, ast.ClassDef):
                field_groups[node.name] = self.visit_field_class(node)
                continue

            if not isinstance(node, ast.Assign):
                continue
            if not isinstance(node.value, ast.Call):
                continue
            if not isinstance(node.targets[0], ast.Name):
                continue

            # Build accessible symbols table.
            symtable = {}
            # All field types.
            symtable.update({field.__name__: field for field in get_available_fields()})
            # Field group classes.
            symtable.update(field_groups)

            evaluator = SafeEvaluator(symtable=symtable)

            name = node.targets[0].id
            try:
                field = evaluator.run(node.value)
            except Exception as ex:
                # TODO: Handle errors.
                raise ex

            if descriptor is not None:
                field.contribute_to_class(descriptor, fields, name)
            else:
                discovered_fields[name] = field

        if descriptor is None:

            class Fields:
                """Fields wrapper."""

            for name, field in discovered_fields.items():
                setattr(Fields, name, field)

            return Fields

    def visit_ClassDef(self, node):
        """Visit top-level classes."""
        # Resolve everything as root scope contains everything from the process module.
        for base in node.bases:
            # Cover `from resolwe.process import ...`.
            if isinstance(base, ast.Name) and isinstance(base.ctx, ast.Load):
                base = python_runtimes_manager.registered_class(base.id)
                if base is not None:
                    self.base_classes.add(base.__name__)

            # Cover `from resolwe import process`.
            elif isinstance(base, ast.Attribute) and isinstance(base.ctx, ast.Load):
                base = python_runtimes_manager.registered_class(base.attr)
                if base is not None:
                    self.base_classes.add(base.__name__)
            else:
                continue

            if isclass(base) and issubclass(base, runtime.Process):
                break
        else:
            return

        descriptor = ProcessDescriptor(source=self.source)
        descriptor.metadata.lineno = node.lineno

        # Available embedded classes.
        embedded_class_fields = {
            runtime.PROCESS_INPUTS_NAME: descriptor.inputs,
            runtime.PROCESS_OUTPUTS_NAME: descriptor.outputs,
        }

        # Parse metadata in class body.
        for item in node.body:
            if isinstance(item, ast.Assign):
                # Possible metadata.
                if (
                    len(item.targets) == 1
                    and isinstance(item.targets[0], ast.Name)
                    and isinstance(item.targets[0].ctx, ast.Store)
                    and item.targets[0].id in PROCESS_METADATA
                ):
                    # Try to get the metadata value.
                    value = PROCESS_METADATA[item.targets[0].id].get_value(item.value)
                    setattr(descriptor.metadata, item.targets[0].id, value)
            elif (
                isinstance(item, ast.Expr)
                and isinstance(item.value, ast.Str)
                and descriptor.metadata.description is None
            ):
                # Possible description string.
                descriptor.metadata.description = item.value.s
            elif (
                isinstance(item, ast.ClassDef)
                and item.name in embedded_class_fields.keys()
            ):
                # Possible input/output declaration.
                self.visit_field_class(
                    item, descriptor, embedded_class_fields[item.name]
                )

        descriptor.validate()
        self.processes.append(descriptor)
        # Add mapping between the slug and the node.
        self.ast_nodes[descriptor.metadata.slug] = node


class SafeParser:
    """Safe parser for Python processes which doesn't evaluate any code."""

    def __init__(self, source):
        """Construct process parser.

        :param source: Process source code string
        """
        self._source = source

    def parse(self):
        """Parse process.

        :return: A list of discovered process descriptors
        """
        root = ast.parse(self._source)
        visitor = ProcessVisitor(source=self._source)
        visitor.visit(root)
        return visitor.processes

    def base_classes(self):
        """Parse process.

        :return: A list of the base classes for the processes.
        """
        root = ast.parse(self._source)
        visitor = ProcessVisitor(source=self._source)
        visitor.visit(root)
        return visitor.base_classes



import os
from io import StringIO

from django.core.management import call_command
from django.db.models import Q
from django.test import TestCase

from resolwe.flow.models import Data, Process
from resolwe.storage.management.commands.compare_models_and_csv import (
    FileIterator,
    ModelIterator,
    map_subpath_locations,
    parse_line,
)
from resolwe.storage.models import FileStorage, ReferencedPath, StorageLocation


class CompareModelsTestCase(TestCase):
    def setUp(self):
        self.proc = Process.objects.create(contributor_id=1)
        # subpath 100
        self.fs100 = FileStorage.objects.create()
        self.d100 = Data.objects.create(
            process=self.proc,
            contributor_id=1,
            location=self.fs100,
        )
        self.sl100 = StorageLocation.objects.create(
            file_storage=self.fs100,
            url="100",
        )
        # subpath 123
        self.fs123 = FileStorage.objects.create()
        self.d123 = Data.objects.create(
            process=self.proc,
            contributor_id=1,
            location=self.fs123,
        )
        self.sl123 = StorageLocation.objects.create(
            file_storage=self.fs123,
            url="123",
        )

        self.make_file = lambda name, sl: ReferencedPath.objects.create(
            path=name, awss3etag="hashhashhash"
        ).storage_locations.add(sl)

        # all paths present in csv:
        self.make_file("names/000.json", self.sl100)
        self.make_file("names/001.json", self.sl100)
        self.make_file("names/002.json", self.sl100)
        self.make_file("names/003.json", self.sl100)
        self.make_file("names/004.json", self.sl100)
        self.make_file("names/005.json", self.sl100)

        self.make_file("d.json", self.sl123)
        self.make_file("f.json", self.sl123)
        self.make_file("j.json", self.sl123)
        self.make_file("k.json", self.sl123)
        self.make_file("l.json", self.sl123)

        self.csv_filename = os.path.abspath(
            os.path.dirname(__file__) + "/files/compare_models_and_csv.csv"
        )

        self.out = StringIO()

    def test_complete_match(self):
        call_command(
            "compare_models_and_csv",
            self.csv_filename,
            stdout=self.out,
        )
        self.output = self.out.getvalue()
        self.assertNotIn("CSV-ONLY", self.output)
        self.assertNotIn("MODEL-ONLY", self.output)
        self.assertIn("11 files OK", self.output)
        self.assertNotIn("files in CSV only", self.output)
        self.assertNotIn("files in models only", self.output)
        self.assertNotIn("Numbers don't add up.", self.output)

    def test_complete_mismatch(self):
        ReferencedPath.objects.all().delete()

        self.make_file("names/006.json", self.sl100)

        self.make_file("a.json", self.sl123)
        self.make_file("b.json", self.sl123)
        self.make_file("c.json", self.sl123)
        self.make_file("e.json", self.sl123)
        self.make_file("g.json", self.sl123)
        self.make_file("h.json", self.sl123)
        self.make_file("i.json", self.sl123)

        call_command(
            "compare_models_and_csv",
            self.csv_filename,
            stdout=self.out,
        )
        self.output = self.out.getvalue()
        self.assertIn("CSV-ONLY 100/names/000.json", self.output)
        self.assertIn("CSV-ONLY 100/names/001.json", self.output)
        self.assertIn("CSV-ONLY 100/names/002.json", self.output)
        self.assertIn("CSV-ONLY 100/names/003.json", self.output)
        self.assertIn("CSV-ONLY 100/names/004.json", self.output)
        self.assertIn("CSV-ONLY 100/names/005.json", self.output)
        self.assertIn("MODEL-ONLY 100/names/006.json", self.output)
        self.assertIn("CSV-ONLY 123/d.json", self.output)
        self.assertIn("CSV-ONLY 123/f.json", self.output)
        self.assertIn("CSV-ONLY 123/j.json", self.output)
        self.assertIn("CSV-ONLY 123/k.json", self.output)
        self.assertIn("CSV-ONLY 123/l.json", self.output)
        self.assertIn("MODEL-ONLY 123/a.json", self.output)
        self.assertIn("MODEL-ONLY 123/b.json", self.output)
        self.assertIn("MODEL-ONLY 123/c.json", self.output)
        self.assertIn("MODEL-ONLY 123/e.json", self.output)
        self.assertIn("MODEL-ONLY 123/g.json", self.output)
        self.assertIn("MODEL-ONLY 123/h.json", self.output)
        self.assertIn("MODEL-ONLY 123/i.json", self.output)
        self.assertIn("0 files OK", self.output)
        self.assertIn("11 files in CSV only", self.output)
        self.assertIn("8 files in models only", self.output)
        self.assertNotIn("Numbers don't add up.", self.output)

    def test_missing_subpath_in_models(self):
        self.sl100.files.all().delete()
        self.d100.delete()

        call_command(
            "compare_models_and_csv",
            self.csv_filename,
            stdout=self.out,
        )
        self.output = self.out.getvalue()
        self.assertIn("CSV-ONLY 100/* (6 files)", self.output)
        self.assertNotIn("MODEL-ONLY", self.output)
        self.assertIn("5 files OK, 6 files in CSV only", self.output)
        self.assertNotIn("files in models only", self.output)
        self.assertNotIn("Numbers don't add up.", self.output)

    def test_missing_subpath_in_csv(self):
        # subpath 200
        self.fs200 = FileStorage.objects.create()
        self.d200 = Data.objects.create(
            process=self.proc,
            contributor_id=1,
            location=self.fs200,
        )
        self.sl200 = StorageLocation.objects.create(
            file_storage=self.fs200,
            url="200",
        )

        self.make_file("x.json", self.sl200)
        self.make_file("y.json", self.sl200)
        self.make_file("z.json", self.sl200)

        call_command(
            "compare_models_and_csv",
            self.csv_filename,
            stdout=self.out,
        )
        self.output = self.out.getvalue()
        self.assertIn("MODEL-ONLY 200/* (3 files)", self.output)
        self.assertNotIn("CSV-ONLY", self.output)
        self.assertIn("11 files OK, 3 files in models only", self.output)
        self.assertNotIn("files in CSV only", self.output)
        self.assertNotIn("Numbers don't add up.", self.output)

    def test_empty_models(self):
        self.sl100.files.all().delete()
        self.d100.delete()
        self.sl123.files.all().delete()
        self.d123.delete()

        call_command(
            "compare_models_and_csv",
            self.csv_filename,
            stdout=self.out,
        )
        self.output = self.out.getvalue()
        self.assertIn("CSV-ONLY 100/* (6 files)", self.output)
        self.assertIn("CSV-ONLY 123/* (5 files)", self.output)
        self.assertNotIn("MODEL-ONLY", self.output)
        self.assertIn("0 files OK, 11 files in CSV only", self.output)
        self.assertNotIn("files in models only", self.output)
        self.assertNotIn("Numbers don't add up.", self.output)

    def test_empty_csv(self):
        call_command("compare_models_and_csv", os.devnull, stdout=self.out)
        self.output = self.out.getvalue()
        self.assertIn("MODEL-ONLY 100/* (6 files)", self.output)
        self.assertIn("MODEL-ONLY 123/* (5 files)", self.output)
        self.assertNotIn("CSV-ONLY", self.output)
        self.assertIn("0 files OK, 11 files in models only", self.output)
        self.assertNotIn("files in CSV only", self.output)
        self.assertNotIn("Numbers don't add up.", self.output)

    def test_missing_first_file_in_models(self):
        ReferencedPath.objects.get(path="names/000.json").delete()
        ReferencedPath.objects.get(path="names/001.json").delete()
        call_command(
            "compare_models_and_csv",
            self.csv_filename,
            stdout=self.out,
        )
        self.output = self.out.getvalue()
        self.assertIn("CSV-ONLY 100/names/000.json", self.output)
        self.assertIn("CSV-ONLY 100/names/001.json", self.output)
        self.assertNotIn("MODEL-ONLY", self.output)
        self.assertIn("9 files OK, 2 files in CSV only", self.output)
        self.assertNotIn("files in models only", self.output)
        self.assertNotIn("Numbers don't add up.", self.output)

    def test_missing_last_file_in_models(self):
        ReferencedPath.objects.get(path="names/004.json").delete()
        ReferencedPath.objects.get(path="names/005.json").delete()
        call_command(
            "compare_models_and_csv",
            self.csv_filename,
            stdout=self.out,
        )
        self.output = self.out.getvalue()
        self.assertIn("CSV-ONLY 100/names/004.json", self.output)
        self.assertIn("CSV-ONLY 100/names/005.json", self.output)
        self.assertNotIn("MODEL-ONLY", self.output)
        self.assertIn("9 files OK, 2 files in CSV only", self.output)
        self.assertNotIn("files in models only", self.output)
        self.assertNotIn("Numbers don't add up.", self.output)

    def test_missing_first_file_in_csv(self):
        self.make_file("a.json", self.sl123)
        self.make_file("b.json", self.sl123)
        self.make_file("c.json", self.sl123)
        call_command(
            "compare_models_and_csv",
            self.csv_filename,
            stdout=self.out,
        )
        self.output = self.out.getvalue()
        self.assertIn("MODEL-ONLY 123/a.json", self.output)
        self.assertIn("MODEL-ONLY 123/b.json", self.output)
        self.assertIn("MODEL-ONLY 123/c.json", self.output)
        self.assertNotIn("CSV-ONLY", self.output)
        self.assertIn("11 files OK, 3 files in models only", self.output)
        self.assertNotIn("files in CSV only", self.output)
        self.assertNotIn("Numbers don't add up.", self.output)

    def test_missing_last_file_in_csv(self):
        self.make_file("names/006.json", self.sl100)
        self.make_file("m.json", self.sl123)
        call_command(
            "compare_models_and_csv",
            self.csv_filename,
            stdout=self.out,
        )
        self.output = self.out.getvalue()
        self.assertIn("MODEL-ONLY 100/names/006.json", self.output)
        self.assertIn("MODEL-ONLY 123/m.json", self.output)
        self.assertNotIn("CSV-ONLY", self.output)
        self.assertIn("11 files OK, 2 files in models only", self.output)
        self.assertNotIn("files in CSV only", self.output)
        self.assertNotIn("Numbers don't add up.", self.output)

    def test_missing_middle_file_in_models(self):
        ReferencedPath.objects.get(path="names/002.json").delete()
        ReferencedPath.objects.get(path="names/003.json").delete()
        call_command(
            "compare_models_and_csv",
            self.csv_filename,
            stdout=self.out,
        )
        self.output = self.out.getvalue()
        self.assertIn("CSV-ONLY 100/names/002.json", self.output)
        self.assertIn("CSV-ONLY 100/names/003.json", self.output)
        self.assertNotIn("MODEL-ONLY", self.output)
        self.assertIn("9 files OK, 2 files in CSV only", self.output)
        self.assertNotIn("files in models only", self.output)
        self.assertNotIn("Numbers don't add up.", self.output)

    def test_missing_middle_file_in_csv(self):
        self.make_file("e.json", self.sl123)
        self.make_file("g.json", self.sl123)
        self.make_file("h.json", self.sl123)
        self.make_file("i.json", self.sl123)
        call_command(
            "compare_models_and_csv",
            self.csv_filename,
            stdout=self.out,
        )
        self.output = self.out.getvalue()
        self.assertIn("MODEL-ONLY 123/e.json", self.output)
        self.assertIn("MODEL-ONLY 123/g.json", self.output)
        self.assertIn("MODEL-ONLY 123/h.json", self.output)
        self.assertIn("MODEL-ONLY 123/i.json", self.output)
        self.assertNotIn("CSV-ONLY", self.output)
        self.assertIn("11 files OK, 4 files in models only", self.output)
        self.assertNotIn("files in CSV only", self.output)
        self.assertNotIn("Numbers don't add up.", self.output)

    def test_partial_mismatch(self):
        # Situation like this:
        # CSV | models
        #  0      0  (start with a match)
        #  1         (models leading)
        #  2         (no common entry for them to meet)
        #         25 (CSV starts to lead)
        #  3
        #         35 (mismatching entries)
        #  4
        #  5      5  (finish with another match)
        self.make_file("names/0025.json", self.sl100)
        self.make_file("names/0035.json", self.sl100)
        ReferencedPath.objects.get(path="names/001.json").delete()
        ReferencedPath.objects.get(path="names/002.json").delete()
        ReferencedPath.objects.get(path="names/003.json").delete()
        ReferencedPath.objects.get(path="names/004.json").delete()
        call_command(
            "compare_models_and_csv",
            self.csv_filename,
            stdout=self.out,
        )
        self.output = self.out.getvalue()
        self.assertIn("MODEL-ONLY 100/names/0025.json", self.output)
        self.assertIn("MODEL-ONLY 100/names/0035.json", self.output)
        self.assertIn("CSV-ONLY 100/names/001.json", self.output)
        self.assertIn("CSV-ONLY 100/names/002.json", self.output)
        self.assertIn("CSV-ONLY 100/names/003.json", self.output)
        self.assertIn("CSV-ONLY 100/names/004.json", self.output)
        self.assertIn("7 files OK", self.output)
        self.assertIn("2 files in models only", self.output)
        self.assertIn("4 files in CSV only", self.output)
        self.assertNotIn("Numbers don't add up.", self.output)

    def test_partial_mismatch_no_common_finish(self):
        # same as test_partial_mismatch, but without the common finish
        self.make_file("names/0025.json", self.sl100)
        self.make_file("names/0035.json", self.sl100)
        ReferencedPath.objects.get(path="names/001.json").delete()
        ReferencedPath.objects.get(path="names/002.json").delete()
        ReferencedPath.objects.get(path="names/003.json").delete()
        ReferencedPath.objects.get(path="names/004.json").delete()
        ReferencedPath.objects.get(path="names/005.json").delete()
        call_command(
            "compare_models_and_csv",
            self.csv_filename,
            stdout=self.out,
        )
        self.output = self.out.getvalue()
        self.assertIn("MODEL-ONLY 100/names/0025.json", self.output)
        self.assertIn("MODEL-ONLY 100/names/0035.json", self.output)
        self.assertIn("CSV-ONLY 100/names/001.json", self.output)
        self.assertIn("CSV-ONLY 100/names/002.json", self.output)
        self.assertIn("CSV-ONLY 100/names/003.json", self.output)
        self.assertIn("CSV-ONLY 100/names/004.json", self.output)
        self.assertIn("CSV-ONLY 100/names/005.json", self.output)
        self.assertIn("6 files OK", self.output)
        self.assertIn("2 files in models only", self.output)
        self.assertIn("5 files in CSV only", self.output)
        self.assertNotIn("Numbers don't add up.", self.output)

    def test_partial_mismatch_no_common_start(self):
        # same as test_partial_mismatch, but without the common beginning
        self.make_file("names/0025.json", self.sl100)
        self.make_file("names/0035.json", self.sl100)
        ReferencedPath.objects.get(path="names/000.json").delete()
        ReferencedPath.objects.get(path="names/001.json").delete()
        ReferencedPath.objects.get(path="names/002.json").delete()
        ReferencedPath.objects.get(path="names/003.json").delete()
        ReferencedPath.objects.get(path="names/004.json").delete()
        call_command(
            "compare_models_and_csv",
            self.csv_filename,
            stdout=self.out,
        )
        self.output = self.out.getvalue()
        self.assertIn("MODEL-ONLY 100/names/0025.json", self.output)
        self.assertIn("MODEL-ONLY 100/names/0035.json", self.output)
        self.assertIn("CSV-ONLY 100/names/000.json", self.output)
        self.assertIn("CSV-ONLY 100/names/001.json", self.output)
        self.assertIn("CSV-ONLY 100/names/002.json", self.output)
        self.assertIn("CSV-ONLY 100/names/003.json", self.output)
        self.assertIn("CSV-ONLY 100/names/004.json", self.output)
        self.assertIn("6 files OK", self.output)
        self.assertIn("2 files in models only", self.output)
        self.assertIn("5 files in CSV only", self.output)
        self.assertNotIn("Numbers don't add up.", self.output)

    def test_common_finish(self):
        # completely mismatched at first, then come together
        ReferencedPath.objects.get(path="names/000.json").delete()
        ReferencedPath.objects.get(path="names/001.json").delete()
        ReferencedPath.objects.get(path="names/002.json").delete()
        ReferencedPath.objects.get(path="names/003.json").delete()
        self.make_file("names/0025.json", self.sl100)
        self.make_file("names/0035.json", self.sl100)
        call_command(
            "compare_models_and_csv",
            self.csv_filename,
            stdout=self.out,
        )
        self.output = self.out.getvalue()
        self.assertIn("MODEL-ONLY 100/names/0025.json", self.output)
        self.assertIn("MODEL-ONLY 100/names/0035.json", self.output)
        self.assertIn("CSV-ONLY 100/names/000.json", self.output)
        self.assertIn("CSV-ONLY 100/names/001.json", self.output)
        self.assertIn("CSV-ONLY 100/names/002.json", self.output)
        self.assertIn("CSV-ONLY 100/names/003.json", self.output)
        self.assertIn("7 files OK", self.output)
        self.assertIn("2 files in models only", self.output)
        self.assertIn("4 files in CSV only", self.output)
        self.assertNotIn("Numbers don't add up.", self.output)

    def test_hash_mismatch(self):
        rp003 = ReferencedPath.objects.get(path="names/003.json")
        rp003.awss3etag = "hashhash"
        rp003.save()
        rp005 = ReferencedPath.objects.get(path="names/005.json")
        rp005.awss3etag = ""
        rp005.save()
        rpd = ReferencedPath.objects.get(path="d.json")
        rpd.awss3etag = "not hash"
        rpd.save()
        call_command(
            "compare_models_and_csv",
            self.csv_filename,
            stdout=self.out,
        )
        self.output = self.out.getvalue()
        longstring = "HASH 100/names/003.json hashhash != hashhashhash"
        self.assertIn(longstring, self.output)  # longstr; 79ch line len limit
        self.assertIn("HASH 100/names/005.json  != hashhashhash", self.output)
        self.assertIn("HASH 123/d.json not hash != hashhashhash", self.output)
        self.assertIn("8 files OK", self.output)
        self.assertNotIn("files in models only", self.output)
        self.assertNotIn("files in CSV only", self.output)
        self.assertIn("3 files do not match the hash", self.output)
        self.assertNotIn("Numbers don't add up.", self.output)

    def test_parse_line(self):
        line = (
            '"bucket","subpath/filename.extension",'
            '"filesize","hashhashhash","STANDARD",""'
        )
        line = parse_line(line)
        self.assertEqual(line.filename, "filename.extension")
        self.assertEqual(line.hash, "hashhashhash")
        self.assertEqual(line.subpath, "subpath")

    def test_parse_line_no_subpath(self):
        line = '"bucket","README.md","filesize","hashhashhash","STANDARD",""'
        line = parse_line(line)
        self.assertEqual(line.filename, "README.md")
        self.assertEqual(line.hash, "hashhashhash")
        self.assertEqual(line.subpath, "")

    def test_parse_line_empty_key(self):
        line = '"bucket","","filesize","hashhashhash","STANDARD",""'
        line = parse_line(line)
        self.assertEqual(line.filename, "")
        self.assertEqual(line.hash, "hashhashhash")
        self.assertEqual(line.subpath, "")

    def test_map_subpath_locations(self):
        fi = FileIterator(self.csv_filename)
        received = map_subpath_locations(fi)
        expected = {
            "100": {"start": 0, "end": 582, "linecount": 6},
            "123": {"start": 582, "end": 1027, "linecount": 5},
        }
        self.assertEqual(received, expected)
        self.assertEqual(fi.length, 11)

    def test_map_subpath_locations_empty_csv(self):
        fi = FileIterator(os.devnull)
        received = map_subpath_locations(fi)
        expected = {}
        self.assertEqual(received, expected)
        self.assertEqual(fi.length, 0)

    def test_fileiterator_tell_seek(self):
        fi = FileIterator(self.csv_filename)
        fi.seek(42)
        self.assertEqual(fi.tell(), 42)

    def test_file_iterator_readline(self):
        fi = FileIterator(self.csv_filename)
        line = (
            '"342286153875-genialis-dev-storage",'
            '"100/names/000.json","filesize","hashhashhash","STANDARD",""\n'
        )
        self.assertEqual(fi.readline(), line)
        self.assertEqual(fi.last_position, 0)

    def test_file_iterator_has_next(self):
        fi = FileIterator(self.csv_filename)
        self.assertTrue(fi.has_next())
        fi.seek(fi.size)
        self.assertFalse(fi.has_next())

    def test_file_iterator_next(self):
        fi = FileIterator(self.csv_filename)
        self.assertEqual(fi.next(), ("names/000.json", "hashhashhash"))

    def test_file_iterator_next_at_eof(self):
        fi = FileIterator(self.csv_filename)
        fi.seek(fi.size)
        self.assertEqual(fi.next(), ("", ""))

    def test_file_iterator_restrict(self):
        fi = FileIterator(self.csv_filename)
        fi.restrict(0, 10)
        self.assertTrue(fi.has_next())
        fi.seek(10)
        self.assertFalse(fi.has_next())

    def test_file_iterator_seek_relative(self):
        fi = FileIterator(self.csv_filename)
        fi.restrict(20, 40)
        fi.seek_relative(10)
        self.assertEqual(fi.tell(), 30)

    def test_model_iterator_next(self):
        urls = self.d123.location.files
        urls = urls.exclude(Q(path__endswith="/"))  # exclude directories
        urls = ModelIterator(urls.order_by("path"))
        self.assertEqual(urls.next(), ("d.json", "hashhashhash"))
        self.assertEqual(urls.next(), ("f.json", "hashhashhash"))
        self.assertEqual(urls.next(), ("j.json", "hashhashhash"))
        self.assertEqual(urls.next(), ("k.json", "hashhashhash"))
        self.assertEqual(urls.next(), ("l.json", "hashhashhash"))
        self.assertEqual(urls.next(), ("", ""))
        self.assertEqual(urls.next(), ("", ""))

    def test_model_iterator_has_next(self):
        urls = self.d123.location.files
        urls = urls.exclude(Q(path__endswith="/"))  # exclude directories
        urls = ModelIterator(urls.order_by("path"))
        self.assertTrue(urls.has_next())
        self.assertEqual(urls.next(), ("d.json", "hashhashhash"))
        self.assertTrue(urls.has_next())
        self.assertEqual(urls.next(), ("f.json", "hashhashhash"))
        self.assertTrue(urls.has_next())
        self.assertEqual(urls.next(), ("j.json", "hashhashhash"))
        self.assertTrue(urls.has_next())
        self.assertEqual(urls.next(), ("k.json", "hashhashhash"))
        self.assertTrue(urls.has_next())
        self.assertEqual(urls.next(), ("l.json", "hashhashhash"))
        self.assertFalse(urls.has_next())



"""Various utils in the observers app."""

import asyncio
from contextlib import suppress
from threading import Event, Thread
from typing import Awaitable, Optional

from asgiref.sync import async_to_sync
from channels.layers import get_channel_layer
from channels.testing import ApplicationCommunicator

from resolwe.observers import consumers
from resolwe.observers.models import BackgroundTask
from resolwe.permissions.utils import assign_contributor_permissions


def start_background_task(
    task_type: consumers.BackgroundTaskType,
    task_description: str,
    task_data: dict,
    request_user,
) -> BackgroundTask:
    """Create the BackgroundTask and start it."""
    task = BackgroundTask.objects.create(description=task_description)
    assign_contributor_permissions(task, request_user)
    packet = {
        "type": task_type.value,
        "task_id": task.id,
        "request_user_id": request_user.id,
        **task_data,
    }
    async_to_sync(get_channel_layer().send)(consumers.BACKGROUND_TASK_CHANNEL, packet)
    return task


class CoroutineInThread(Thread):
    """Start the coroutine in a new thread."""

    def __init__(self, coroutine: Awaitable):
        """Store a coroutine for later use."""
        super().__init__()
        self._coroutine = coroutine
        self._asyncio_task: Optional[asyncio.Task] = None
        self._loop: Optional[asyncio.AbstractEventLoop] = None

    async def wrapper(self):
        """Start the given coroutine and waits for its completion."""
        # The event must be initalized in the loop.
        self._loop = asyncio.get_running_loop()
        self._asyncio_task = asyncio.create_task(self._coroutine)
        with suppress(asyncio.CancelledError):
            await self._asyncio_task

    def run(self):
        """Start the new event loop."""
        asyncio.run(self.wrapper())

    def stop(self):
        """Stop the thread.

        The method blocks until the thread is stopped.
        """
        if self._asyncio_task is not None and self._loop is not None:
            self._loop.call_soon_threadsafe(self._asyncio_task.cancel)
            self.join()


class BackgroundTaskConsumerManager:
    """Start the background task consumer.

    The class can be used as a sync or async context manager. The task can be started
    inside given event loop or in a new thread.
    """

    def __init__(self):
        """Initialize the task attribute."""
        self._thread: Optional[CoroutineInThread] = None
        self._loop: Optional[asyncio.AbstractEventLoop] = None
        self._task: Optional[asyncio.Task] = None
        self._coroutine_stopped: Event = Event()

    def set_loop(self, loop: Optional[asyncio.AbstractEventLoop]):
        """Set the event loop to run manager in.

        When loop is set to None, the manager is started in a new loop in a separate thread.
        """
        self._loop = loop

    def start_task(self):
        """Start the consumer task.

        When loop is set the task is started in the loop. Otherwise, the task is started
        in a new thread.

        If task is already started, the method has no effect.
        """
        if self._loop is not None:
            self._task = self._loop.create_task(self._listen_for_messages())

        elif self._thread is None:
            self._thread = CoroutineInThread(self._listen_for_messages())
            self._thread.start()

    def stop_task(self):
        """Stop the consumer task.

        When task is not started, the method has no effect.
        """
        if self._loop is not None and self._task is not None:
            self._loop.call_soon_threadsafe(self._task.cancel)
            self._coroutine_stopped.wait()
            self._coroutine_stopped.clear()
        elif self._thread is not None:
            self._thread.stop()
            self._thread = None

    async def _listen_for_messages(self):
        """Listen to messages and forward them to the consumer."""
        try:
            scope = {"channel": consumers.BACKGROUND_TASK_CHANNEL}
            app = ApplicationCommunicator(consumers.BackgroundTaskConsumer(), scope)
            channel_layer = get_channel_layer()
            while True:
                message = await channel_layer.receive(consumers.BACKGROUND_TASK_CHANNEL)
                await app.send_input({**message, **scope})
        finally:
            app.stop()
            self._coroutine_stopped.set()

    def __enter__(self):
        """Start the consumer task."""
        self.start_task()

    async def __aenter__(self):
        """Start the consumer task."""
        self.__enter__()

    def __exit__(self, *args, **kwargs):
        """Stop the consumer task."""
        self.stop_task()

    async def __aexit__(self, *args, **kwargs):
        """Stop the consumer task."""
        # Stop the task if it is running in the current loop.
        if self._loop == asyncio.get_running_loop() and self._task is not None:
            self._task.cancel()
            with suppress(asyncio.CancelledError):
                await self._task
        # Stop the task running in loop in a separate thread.
        else:
            self.__exit__(*args, **kwargs)


background_task_manager = BackgroundTaskConsumerManager()



#!/usr/bin/env python3
"""Utility to reformat cURL progress to process progress."""
import argparse
import sys

parser = argparse.ArgumentParser(
    description="Reformat cURL progress to process progress."
)
parser.add_argument(
    "--scale", type=float, default=1.0, help="max progress at the end of transfer"
)
parser.add_argument(
    "--nprints", type=int, default=100, help="number of progress reports"
)
args = parser.parse_args()

nprints = args.nprints
scale = args.scale
step = scale / nprints
milestone = step

while True:
    line = sys.stdin.readline()

    if not line:
        break

    perc = float(line) * scale / 100.0

    if perc >= milestone:
        print('{{"proc.progress":{}}}'.format(perc))
        milestone += step

print('{{"proc.progress":{}}}'.format(scale))



from resolwe.process import Process, StringField


class PythonProcessExample(Process):
    """This process is used for testing of autoprocess Sphinx directive."""

    slug = "test-autoprocess-python"
    name = "Test autoprocess Python"
    version = "1.0.0"
    process_type = "data:python:autoprocess"

    class Input:
        """Input fields."""

        foo = StringField(label="Foo")

    class Output:
        foo = StringField(label="Foo")

    def run(self, inputs, outputs):
        outputs.foo = inputs.foo



""".. Ignore pydocstyle D400.

==================
List Docker images
==================

"""

import functools
import logging
import operator
import shlex
import subprocess
import threading
import time

import yaml
from django.conf import settings
from django.core.management.base import BaseCommand, CommandError

from resolwe.flow.models import Process

logger = logging.getLogger(__name__)

# Global list of images already pulled when running in the same process. This
# is used to avoid pulling images more than once when executed from tests.
PULLED_IMAGES = set()
PULLED_IMAGES_LOCK = threading.Lock()


class Command(BaseCommand):
    """List Docker images used by processes.  Optionally also pull them."""

    help = "List Docker images used by processes in either plain text or YAML (for Ansible)"

    def add_arguments(self, parser):
        """Add an argument to specify output format."""
        parser.add_argument(
            "--format",
            dest="format",
            default="plain",
            help="Set output format ('plain' [default] or 'yaml')",
        )
        parser.add_argument(
            "--pull",
            dest="pull",
            default=False,
            action="store_true",
            help="Pull all images with Docker",
        )
        parser.add_argument(
            "--ignore-pull-errors",
            dest="ignore_pull_errors",
            default=getattr(settings, "FLOW_DOCKER_IGNORE_PULL_ERRORS", False),
            action="store_true",
            help="Don't fail whenever a Docker image can't be pulled",
        )

    def _map_image(self, docker_image: str, mapper: dict[str, str]) -> str:
        """Transform the given Docker image according to the mapper.

        If image starts with the prefix in the mapper keys, it is replaced with the
        corresponding value. If no prefix matches, the image is returned unchanged.
        """
        for prefix, replacement in mapper.items():
            if docker_image.startswith(prefix):
                return replacement + docker_image[len(prefix) :]
        return docker_image

    def _map_images(self, docker_images: set[str]) -> set[str]:
        """Map the docker images in the given set according to the settings."""
        mapper = getattr(settings, "FLOW_CONTAINER_IMAGE_MAP", {})
        return {self._map_image(image, mapper) for image in docker_images}

    def handle(self, *args, **options):
        """Handle command list_docker_images."""
        verbosity = int(options.get("verbosity"))

        # Check that the specified output format is valid
        if options["format"] != "plain" and options["format"] != "yaml":
            raise CommandError("Unknown output format: %s" % options["format"])

        # Gather only unique latest custom Docker requirements that the processes are using
        # The 'image' field is optional, so be careful about that as well
        unique_docker_images = set(
            p.requirements["executor"]["docker"]["image"]
            for p in Process.objects.filter(is_active=True)
            .only("requirements")
            .filter(requirements__icontains="docker")
            if "image" in p.requirements.get("executor", {}).get("docker", {})
        )

        # Add the default image.
        unique_docker_images.add(
            getattr(
                settings,
                "FLOW_DOCKER_DEFAULT_PROCESSING_CONTAINER_IMAGE",
                "public.ecr.aws/s4q6j6e8/resolwe/base:ubuntu-20.04",
            )
        )

        # Map the Docker images according to the settings.
        unique_docker_images = self._map_images(unique_docker_images)

        # Pull images if requested or just output the list in specified format
        if options["pull"]:
            # Remove set of already pulled images.
            with PULLED_IMAGES_LOCK:
                unique_docker_images.difference_update(PULLED_IMAGES)

            # Get the desired 'docker' command from settings or use the default
            docker = getattr(settings, "FLOW_DOCKER_COMMAND", "docker")

            # Pull each image
            for img in unique_docker_images:
                max_retries = 10
                for _ in range(max_retries):
                    ret = subprocess.run(
                        shlex.split("{} pull {}".format(docker, img)),
                        stdout=subprocess.PIPE,
                        stderr=subprocess.STDOUT,
                    )
                    if ret.returncode == 0:
                        break
                    time.sleep(1)

                # Update set of pulled images.
                with PULLED_IMAGES_LOCK:
                    PULLED_IMAGES.add(img)

                if ret.returncode != 0:
                    errmsg = "Failed to pull Docker image '{}': '{}'.".format(
                        img, ret.stdout.decode()
                    )

                    if not options["ignore_pull_errors"]:
                        # Print error and stop execution
                        raise CommandError(errmsg)
                    else:
                        # Print error, but keep going
                        logger.error(errmsg)
                        if verbosity > 0:
                            self.stderr.write(errmsg)
                else:
                    msg = "Docker image '{}' pulled successfully!".format(img)
                    logger.info(msg)
                    if verbosity > 0:
                        self.stdout.write(msg)
        else:
            # Sort the set of unique Docker images for nicer output.
            unique_docker_images = sorted(unique_docker_images)

            # Convert the set of unique Docker images into a list of dicts for easier output
            imgs = [
                dict(name=s[0], tag=s[1] if len(s) == 2 else "latest")
                for s in (img.split(":") for img in unique_docker_images)
            ]

            # Output in YAML or plaintext (one image per line), as requested
            if options["format"] == "yaml":
                out = yaml.safe_dump(imgs, default_flow_style=True, default_style="'")
            else:
                out = functools.reduce(
                    operator.add, ("{name}:{tag}\n".format(**i) for i in imgs), ""
                )

            self.stdout.write(out, ending="")



"""Channels consumers for Storage application."""

import logging

from channels.consumer import SyncConsumer

from resolwe.storage.cleanup import Cleaner
from resolwe.storage.manager import Manager

logger = logging.getLogger(__name__)


class StorageManagerConsumer(SyncConsumer):
    """Start Storage Manager when triggered."""

    def storagemanager_run(self, event):
        """Start the manager run."""
        manager = Manager()
        try:
            manager.process()
        except Exception:
            logger.exception("Error while running manager.")


class StorageCleanupConsumer(SyncConsumer):
    """Start Storage Cleanup when triggered.

    Optionally id of the FileStorage object to clean can be sent with the
    event.
    """

    def storagecleanup_run(self, event):
        """Start the cleanup run."""
        cleaner = Cleaner()
        try:
            file_storage_id = event.get("file_storage_id")
            cleaner.process(file_storage_id)
        except Exception:
            logger.exception("Error while running cleanup.")



"""Mixins used in Resolwe Viewsets."""

from django.db import IntegrityError, transaction
from drf_spectacular.utils import extend_schema
from rest_framework import exceptions, mixins, serializers, status
from rest_framework.decorators import action
from rest_framework.generics import get_object_or_404
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response

from resolwe.observers.models import BackgroundTask
from resolwe.observers.views import BackgroundTaskSerializer
from resolwe.permissions.models import (
    Permission,
    PermissionInterface,
    get_anonymous_user,
)
from resolwe.permissions.utils import assign_contributor_permissions


class DuplicateSerializer(serializers.Serializer):
    """Deserializer for duplicate endpoint."""

    ids = serializers.ListField(child=serializers.IntegerField(), allow_empty=False)


class DeleteSerializer(serializers.Serializer):
    """Bulk delete given objects."""

    ids = serializers.ListField(child=serializers.IntegerField(), allow_empty=False)


class ResolweCreateModelMixin(mixins.CreateModelMixin):
    """Mixin to support creating new `Resolwe` models.

    Extends `django_rest_framework`'s class `CreateModelMixin` with:

      * append user's id from request to posted data as `contributor`
        key
      * catch `IntegrityError`s, so we can return HTTP status 409
        instead of raising error

    """

    def resolve_user(self, user):
        """Resolve user instance from request."""
        return get_anonymous_user() if user.is_anonymous else user

    def define_contributor(self, request):
        """Define contributor by adding it to request.data.

        When data is a list, iterate over it and add the contributor to each entry.
        """
        contributor_pk = self.resolve_user(request.user).pk
        for entry in request.data if isinstance(request.data, list) else [request.data]:
            entry["contributor"] = contributor_pk

    def create(self, request, *args, **kwargs):
        """Create a resource."""
        self.define_contributor(request)
        try:
            return super().create(request, *args, **kwargs)
        except IntegrityError as ex:
            return Response({"error": str(ex)}, status=status.HTTP_409_CONFLICT)

    def perform_create(self, serializer):
        """Create a resource."""
        with transaction.atomic():
            instance = serializer.save()
            if isinstance(instance, PermissionInterface):
                # Assign all permissions to the object contributor when object is not
                # in container.
                if not instance.in_container():
                    assign_contributor_permissions(instance)


class ResolweBackgroundDuplicateMixin:
    """Duplicate objects in the background."""

    # The method background queryset must have.
    BACKGROUND_DUPLICATE_METHOD = "duplicate"

    @extend_schema(
        filters=False,
        request=DuplicateSerializer(),
        responses={200: BackgroundTaskSerializer},
    )
    @action(detail=False, methods=["post"], permission_classes=[IsAuthenticated])
    def duplicate(self, request, *args, **kwargs):
        """Bulk duplicate  objects in the background.

        The background task is returned.
        """
        serializer = DuplicateSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        Model = self.serializer_class.Meta.model

        # First check if the queryset supports background duplication.
        if not hasattr(Model.objects.none(), self.BACKGROUND_DUPLICATE_METHOD):
            raise exceptions.ValidationError(
                f"The model '{Model}' does not support background duplication."
            )

        ids = serializer.validated_data["ids"]
        to_duplicate = Model.objects.filter(id__in=ids).filter_for_user(request.user)
        actual_ids = to_duplicate.values_list("id", flat=True)
        missing_ids = list(set(ids) - set(actual_ids))
        if missing_ids:
            id_string = ", ".join(map(str, missing_ids))
            raise exceptions.PermissionDenied(
                f"Objects with ids {id_string} not found."
            )

        duplicate_method = getattr(to_duplicate, self.BACKGROUND_DUPLICATE_METHOD)
        task = duplicate_method(request_user=request.user)

        return Response(
            status=status.HTTP_200_OK,
            data=BackgroundTaskSerializer(task).data,
        )


class ResolweBackgroundDeleteMixin(mixins.DestroyModelMixin):
    """Delete a model instance in the background if possible."""

    BACKGROUND_DELETE_METHOD = "delete_background"

    def perform_destroy(self, instance):
        """Perform the actual delete.

        If possible, delete the object in the background.
        """
        return getattr(instance, self.BACKGROUND_DELETE_METHOD, "delete")()

    def destroy(self, request, *args, **kwargs):
        """Delete a resource."""
        instance = self.get_object()
        return_value = self.perform_destroy(instance)
        # When the object is deleted in the background, return a task description.
        if isinstance(return_value, BackgroundTask):
            return Response(
                status=status.HTTP_200_OK,
                data=BackgroundTaskSerializer(return_value).data,
            )
        else:
            return Response(status=status.HTTP_204_NO_CONTENT)

    @extend_schema(
        filters=False,
        request=DeleteSerializer(),
        responses={200: BackgroundTaskSerializer},
    )
    @action(detail=False, methods=["post"])
    def bulk_delete(self, request, *args, **kwargs):
        """Bulk delete  objects in the background.

        The background task is returned.
        """
        serializer = DeleteSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        ids = set(serializer.validated_data["ids"])

        # The list of ids must not be empty.
        if not ids:
            raise exceptions.ValidationError("No ids provided.")

        queryset = self.get_queryset()
        if not getattr(queryset, self.BACKGROUND_DELETE_METHOD, None):
            raise exceptions.ValidationError(
                f"Model {queryset.model} does not support background deletion."
            )

        can_view = queryset.filter(id__in=ids).filter_for_user(
            request.user, Permission.VIEW
        )
        can_delete = can_view.filter_for_user(request.user, Permission.EDIT)
        can_view_ids = set(can_view.values_list("id", flat=True))
        can_delete_ids = set(can_delete.values_list("id", flat=True))

        if missing_ids := can_view_ids - can_delete_ids:
            joined_ids = ", ".join(map(str, missing_ids))
            raise exceptions.PermissionDenied(
                f"No permission to delete objects with ids {joined_ids}."
            )

        # User can not delete any object.
        if not can_delete_ids:
            raise exceptions.PermissionDenied(
                f"No permission to delete objects with ids {ids}."
            )

        # Start the background task.
        task = getattr(can_delete, self.BACKGROUND_DELETE_METHOD, None)(request.user)
        return Response(
            status=status.HTTP_200_OK,
            data=BackgroundTaskSerializer(task).data,
        )


class ResolweUpdateModelMixin(mixins.UpdateModelMixin):
    """Mixin to support updating `Resolwe` models.

    Extends `django_rest_framework`'s class `UpdateModelMixin` with:
    """

    def update(self, request, *args, **kwargs):
        """Update a resource."""
        # NOTE: Use the original method instead when support for locking is added:
        #       https://github.com/encode/django-rest-framework/issues/4675
        # return super().update(request, *args, **kwargs)
        with transaction.atomic():
            return self._update(request, *args, **kwargs)

    # NOTE: This is a copy of rest_framework.mixins.UpdateModelMixin.update().
    #       The self.get_object() was replaced with the
    #       self.get_object_with_lock() to lock the object while updating.
    #       Use the original method when suport for locking is added:
    #       https://github.com/encode/django-rest-framework/issues/4675
    def _update(self, request, *args, **kwargs):
        """Update a resource."""
        partial = kwargs.pop("partial", False)
        # NOTE: The line below was changed.
        instance = self.get_object_with_lock()
        serializer = self.get_serializer(instance, data=request.data, partial=partial)
        serializer.is_valid(raise_exception=True)
        self.perform_update(serializer)

        if getattr(instance, "_prefetched_objects_cache", None):
            # If 'prefetch_related' has been applied to a queryset, we need to
            # forcibly invalidate the prefetch cache on the instance.
            instance._prefetched_objects_cache = {}

        return Response(serializer.data)

    # NOTE: This is a copy of rest_framework.generics.GenericAPIView.get_object().
    #       The select_for_update() was added to the 'queryset'
    #       to lock the object while updating.
    #       Use the original method when suport for locking is added:
    #       https://github.com/encode/django-rest-framework/issues/4675
    def get_object_with_lock(self):
        """Return the object the view is displaying."""
        queryset = self.filter_queryset(self.get_queryset())

        # Perform the lookup filtering.
        lookup_url_kwarg = self.lookup_url_kwarg or self.lookup_field

        assert lookup_url_kwarg in self.kwargs, (
            "Expected view %s to be called with a URL keyword argument "
            'named "%s". Fix your URL conf, or set the `.lookup_field` '
            "attribute on the view correctly."
            % (self.__class__.__name__, lookup_url_kwarg)
        )

        filter_kwargs = {self.lookup_field: self.kwargs[lookup_url_kwarg]}
        # NOTE: The line below was changed.
        obj = get_object_or_404(queryset.select_for_update(), **filter_kwargs)

        # May raise a permission denied.
        self.check_object_permissions(self.request, obj)

        return obj


class NameSerializer(serializers.Serializer):
    """Serializer for the name field."""

    name = serializers.CharField()


class ResolweCheckSlugMixin:
    """Slug validation."""

    @extend_schema(parameters=[NameSerializer()], responses={status.HTTP_200_OK: bool})
    @action(detail=False, methods=["get"])
    def slug_exists(self, request):
        """Check if given url slug exists.

        Check if slug given in query parameter ``name`` exists. Return
        ``True`` if slug already exists and ``False`` otherwise.

        """
        if not request.user.is_authenticated:
            return Response(status=status.HTTP_401_UNAUTHORIZED)

        serializer = NameSerializer(data=request.query_params)
        serializer.is_valid(raise_exception=True)
        slug_name = serializer.validated_data["name"]

        queryset = self.get_queryset()
        return Response(queryset.filter(slug__iexact=slug_name).exists())



# pylint: disable=missing-docstring

import asyncio
import json
import uuid

from channels.db import database_sync_to_async
from channels.routing import URLRouter
from channels.testing import WebsocketCommunicator
from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType
from django.test import TransactionTestCase, override_settings
from django.urls import path
from rest_framework import status
from rest_framework.test import APIRequestFactory, force_authenticate

from resolwe.flow.models import Collection, Data, DescriptorSchema, Entity, Process
from resolwe.flow.views import DataViewSet, EntityViewSet
from resolwe.permissions.models import Permission
from resolwe.permissions.utils import get_anonymous_user
from resolwe.test import TransactionResolweAPITestCase

from .consumers import ClientConsumer
from .models import BackgroundTask, Observer, Subscription
from .protocol import ChangeType
from .views import BackgroundTaksViewSet, BackgroundTaskSerializer


# If FLOW_MANAGER_DISABLE_AUTO_CALLS is False, Data objects will receive
# an UPDATE signal before the CREATE signal.
@override_settings(FLOW_MANAGER_DISABLE_AUTO_CALLS=True)
class ObserverTestCase(TransactionTestCase):
    def setUp(self):
        super().setUp()

        self.user_alice = get_user_model().objects.create(
            username="alice",
            email="alice@test.com",
            first_name="Ana",
            last_name="Banana",
        )
        self.user_bob = get_user_model().objects.create(
            username="capital-bob",
            email="bob@bob.bob",
            first_name="Capital",
            last_name="Bobnik",
        )
        DescriptorSchema.objects.create(
            name="Sample", slug="sample", contributor=self.user_alice
        )
        self.process = Process.objects.create(
            name="Dummy process", contributor=self.user_alice
        )

        self.process_auto_entity = Process.objects.create(
            name="Dummy process",
            contributor=self.user_alice,
            entity_type="sample",
        )

        self.client_consumer = URLRouter(
            [path("ws/<slug:session_id>", ClientConsumer().as_asgi())]
        )
        self.subscription_id = uuid.UUID(int=0)
        self.subscription_id2 = uuid.UUID(int=1)

    async def assert_no_more_messages(self, client):
        """Assert there are no messages queued by a websocket client."""
        with self.assertRaises(asyncio.TimeoutError):
            async with asyncio.timeout(0.01):
                raise ValueError("Unexpected message:", await client.receive_from())

    async def await_subscription_observer_count(self, count):
        """Wait until the number of all subscriptions is equal to count."""

        @database_sync_to_async
        def get_subscription_count():
            subs = Subscription.objects.filter(session_id="test_session")
            total = 0
            for sub in subs:
                total += sub.observers.count()
            return total

        try:
            async with asyncio.timeout(1):
                while await get_subscription_count() != count:
                    await asyncio.sleep(0.01)
        except asyncio.TimeoutError:
            raise ValueError(
                "Expecting subscription-observer count to be {count}, but it's {actual}".format(
                    count=count, actual=await get_subscription_count()
                )
            )

    async def await_object_count(self, object, count):
        """Wait until the number of all observers is equal to count."""

        @database_sync_to_async
        def get_count():
            return object.objects.count()

        try:
            async with asyncio.timeout(1):
                while await get_count() != count:
                    await asyncio.sleep(0.01)
        except asyncio.TimeoutError:
            raise ValueError(
                "Expecting {obj} count to be {count}, but it's {actual}".format(
                    obj=object.__name__, count=count, actual=await get_count()
                )
            )

    async def websocket_subscribe_unsubscribe(self, Object, create_obj):
        client = WebsocketCommunicator(self.client_consumer, "/ws/test_session")
        connected, _ = await client.connect()
        self.assertTrue(connected)
        await self.await_subscription_observer_count(0)

        # Subscribe to C/D and U separately.
        @database_sync_to_async
        def subscribe():
            content_type = ContentType.objects.get_for_model(Object)
            Subscription.objects.create(
                user=self.user_alice,
                session_id="test_session",
                subscription_id=self.subscription_id,
            ).subscribe(
                content_type=content_type,
                object_ids=[43],
                change_types=[ChangeType.CREATE, ChangeType.DELETE],
            )
            Subscription.objects.create(
                user=self.user_alice,
                session_id="test_session",
                subscription_id=self.subscription_id2,
            ).subscribe(
                content_type=content_type,
                object_ids=[43],
                change_types=[ChangeType.UPDATE],
            )
            return content_type

        content_type = await subscribe()

        await self.await_object_count(Subscription, 2)
        await self.await_object_count(Observer, 3)
        await self.await_subscription_observer_count(3)

        # Create an object.
        obj = await create_obj(pk=43)
        await database_sync_to_async(obj.set_permission)(
            Permission.OWNER, self.user_alice
        )

        # Assert we detect creations.
        self.assertDictEqual(
            json.loads(await client.receive_from()),
            {
                "object_id": 43,
                "change_type": ChangeType.CREATE.name,
                "subscription_id": self.subscription_id.hex,
                "source": [content_type.name, 43],
            },
        )

        # Assert we detect modifications.
        obj.name = "name2"
        await database_sync_to_async(obj.save)()
        self.assertDictEqual(
            json.loads(await client.receive_from()),
            {
                "object_id": 43,
                "change_type": ChangeType.UPDATE.name,
                "subscription_id": self.subscription_id2.hex,
                "source": [content_type.name, 43],
            },
        )

        # Unsubscribe from updates.
        @database_sync_to_async
        def unsubscribe():
            Subscription.objects.get(
                observers__object_id=43, observers__change_type=ChangeType.UPDATE.value
            ).delete()

        await unsubscribe()

        await self.await_object_count(Subscription, 1)
        # Update observer should be deleted because no one is subscribed to it.
        await self.await_object_count(Observer, 2)
        await self.await_subscription_observer_count(2)

        # Assert we don't detect updates anymore.
        obj.name = "name2"
        await database_sync_to_async(obj.save)()
        await self.assert_no_more_messages(client)

        # Assert we detect deletions.
        await database_sync_to_async(obj.delete)()
        self.assertDictEqual(
            json.loads(await client.receive_from()),
            {
                "object_id": 43,
                "change_type": ChangeType.DELETE.name,
                "subscription_id": self.subscription_id.hex,
                "source": [content_type.name, 43],
            },
        )

        await client.disconnect()

    async def test_websocket_subscribe_unsubscribe_collection(self):
        @database_sync_to_async
        def create_collection(pk):
            return Collection.objects.create(
                pk=pk,
                name="Test collection",
                slug="test-collection",
                contributor=self.user_alice,
            )

        await self.websocket_subscribe_unsubscribe(Collection, create_collection)

    async def test_websocket_subscribe_unsubscribe_data(self):
        @database_sync_to_async
        def create_data(pk):
            data = Data.objects.create(
                pk=pk,
                name="Test data",
                slug="test-data",
                contributor=self.user_alice,
                process=self.process,
                size=0,
            )
            return data

        await self.websocket_subscribe_unsubscribe(Data, create_data)

    async def test_websocket_subscribe_unsubscribe_entity(self):
        @database_sync_to_async
        def create_entity(pk):
            return Entity.objects.create(
                pk=pk,
                name="Test entity",
                slug="test-entity",
                contributor=self.user_alice,
            )

        await self.websocket_subscribe_unsubscribe(Entity, create_entity)

    async def test_remove_observers_after_socket_close(self):
        client = WebsocketCommunicator(self.client_consumer, "/ws/test_session")
        client.scope["user"] = self.user_alice
        connected, _ = await client.connect()
        self.assertTrue(connected)

        @database_sync_to_async
        def subscribe():
            Subscription.objects.create(
                user=self.user_alice,
                session_id="test_session",
                subscription_id=self.subscription_id,
            ).subscribe(
                content_type=ContentType.objects.get_for_model(Data),
                object_ids=[42],
                change_types=[ChangeType.UPDATE],
            )

        await subscribe()
        await self.await_subscription_observer_count(1)
        await client.disconnect()
        await self.await_subscription_observer_count(0)

    async def test_observe_containers_update(self):
        """Test containers are also observed when object in them changes."""
        client = WebsocketCommunicator(self.client_consumer, "/ws/test_session")
        connected, _ = await client.connect()
        self.assertTrue(connected)

        @database_sync_to_async
        def create_entity_collection():
            collection = Collection.objects.create(
                pk=40, contributor=self.user_alice, name="test collection"
            )
            entity = Entity.objects.create(
                pk=41,
                contributor=self.user_alice,
                name="test entity",
                collection=collection,
            )
            data = Data.objects.create(
                pk=42,
                name="Test data",
                slug="test-data",
                contributor=self.user_alice,
                process=self.process,
                size=0,
                collection=collection,
                entity=entity,
            )
            collection.set_permission(Permission.OWNER, self.user_alice)
            return collection, entity, data

        # Create collection, entity and data before subscribing.
        collection, entity, data = await create_entity_collection()

        # Create a subscription to the Data, Entity and Collection content_type.
        @database_sync_to_async
        def subscribe():
            subscription = Subscription.objects.create(
                user=self.user_alice,
                session_id="test_session",
                subscription_id=self.subscription_id,
            )

            subscription.subscribe(
                content_type=ContentType.objects.get_for_model(Data),
                object_ids=[data.id],
                change_types=[ChangeType.UPDATE],
            )
            subscription.subscribe(
                content_type=ContentType.objects.get_for_model(Entity),
                object_ids=[entity.id],
                change_types=[ChangeType.UPDATE],
            )
            subscription.subscribe(
                content_type=ContentType.objects.get_for_model(Collection),
                object_ids=[collection.id],
                change_types=[ChangeType.UPDATE],
            )

        await subscribe()
        await self.await_subscription_observer_count(3)

        # Create a new Data object in the collection and entity.
        @database_sync_to_async
        def update_data():
            data.status = Data.STATUS_DONE
            data.save()

        await update_data()

        updates = [json.loads(await client.receive_from()) for _ in range(3)]
        self.assertCountEqual(
            updates,
            [
                {
                    "change_type": ChangeType.UPDATE.name,
                    "object_id": 40,
                    "subscription_id": self.subscription_id.hex,
                    "source": ["data", 42],
                },
                {
                    "change_type": ChangeType.UPDATE.name,
                    "object_id": 41,
                    "subscription_id": self.subscription_id.hex,
                    "source": ["data", 42],
                },
                {
                    "change_type": ChangeType.UPDATE.name,
                    "object_id": 42,
                    "subscription_id": self.subscription_id.hex,
                    "source": ["data", 42],
                },
            ],
        )
        await self.assert_no_more_messages(client)

        @database_sync_to_async
        def update_entity():
            entity.name = "New entity name"
            entity.save()

        await update_entity()
        updates = [json.loads(await client.receive_from()) for _ in range(2)]
        self.assertCountEqual(
            updates,
            [
                {
                    "change_type": ChangeType.UPDATE.name,
                    "object_id": 40,
                    "subscription_id": self.subscription_id.hex,
                    "source": ["entity", 41],
                },
                {
                    "change_type": ChangeType.UPDATE.name,
                    "object_id": 41,
                    "subscription_id": self.subscription_id.hex,
                    "source": ["entity", 41],
                },
            ],
        )
        await self.assert_no_more_messages(client)

    async def test_observe_containers_create(self):
        """Test containers are also observed when object is added."""
        client = WebsocketCommunicator(self.client_consumer, "/ws/test_session")
        connected, _ = await client.connect()
        self.assertTrue(connected)

        @database_sync_to_async
        def create_collection():
            collection = Collection.objects.create(
                pk=40, contributor=self.user_alice, name="test collection"
            )
            collection.set_permission(Permission.OWNER, self.user_alice)
            return collection

        # Create collection before subscribing.
        collection = await create_collection()

        # Create a subscription to Collection object.
        @database_sync_to_async
        def subscribe():
            subscription = Subscription.objects.create(
                user=self.user_alice,
                session_id="test_session",
                subscription_id=self.subscription_id,
            )
            subscription.subscribe(
                content_type=ContentType.objects.get_for_model(Collection),
                object_ids=[collection.id],
                change_types=[ChangeType.CREATE],
            )

        await subscribe()
        await self.await_subscription_observer_count(1)

        # Create a new Entity object in the collection.
        @database_sync_to_async
        def create_entity():
            factory = APIRequestFactory()
            entity = {
                "name": "Entity name",
                "collection": {"id": collection.id},
            }
            request = factory.post("/", entity, format="json")
            force_authenticate(request, self.user_alice)
            viewset = EntityViewSet.as_view(actions={"post": "create"})
            response = viewset(request)
            return Entity.objects.get(pk=response.data["id"])

        entity = await create_entity()
        update = json.loads(await client.receive_from())
        self.assertDictEqual(
            update,
            {
                "change_type": ChangeType.CREATE.name,
                "object_id": collection.id,
                "subscription_id": self.subscription_id.hex,
                "source": ["entity", entity.id],
            },
        )
        await self.assert_no_more_messages(client)

        # Create a subscription to entity.
        @database_sync_to_async
        def subscribe_entity():
            entity = Entity.objects.get()

            subscription = Subscription.objects.get(
                user=self.user_alice,
                session_id="test_session",
                subscription_id=self.subscription_id,
            )
            subscription.subscribe(
                content_type=ContentType.objects.get_for_model(Entity),
                object_ids=[entity.id],
                change_types=[ChangeType.CREATE],
            )

        await subscribe_entity()

        @database_sync_to_async
        def create_data():
            self.process.set_permission(Permission.OWNER, self.user_alice)
            factory = APIRequestFactory()
            data = {
                "process": {"slug": self.process.slug},
                "collection": {"id": collection.id},
                "entity": {"id": entity.id},
            }
            request = factory.post("/", data, format="json")
            force_authenticate(request, self.user_alice)
            viewset = DataViewSet.as_view(actions={"post": "create"})
            response = viewset(request)
            return response.data["id"]

        data_id = await create_data()
        updates = [json.loads(await client.receive_from()) for _ in range(2)]
        self.assertCountEqual(
            updates,
            [
                {
                    "change_type": ChangeType.CREATE.name,
                    "object_id": collection.id,
                    "subscription_id": self.subscription_id.hex,
                    "source": ["data", data_id],
                },
                {
                    "change_type": ChangeType.CREATE.name,
                    "object_id": entity.id,
                    "subscription_id": self.subscription_id.hex,
                    "source": ["data", data_id],
                },
            ],
        )
        await self.assert_no_more_messages(client)

    async def test_observe_data_table(self):
        """Assert creation and deletion messages are sent on data table."""
        client = WebsocketCommunicator(self.client_consumer, "/ws/test_session")
        connected, _ = await client.connect()
        self.assertTrue(connected)

        # Create a subscription to the Data content_type.
        @database_sync_to_async
        def subscribe():
            Subscription.objects.create(
                user=self.user_alice,
                session_id="test_session",
                subscription_id=self.subscription_id,
            ).subscribe(
                content_type=ContentType.objects.get_for_model(Data),
                object_ids=[Observer.ALL_IDS],
                change_types=[ChangeType.CREATE, ChangeType.DELETE],
            )

        await subscribe()
        await self.await_subscription_observer_count(2)

        # Create a new Data object.
        @database_sync_to_async
        def create_data():
            return Data.objects.create(
                name="Test data",
                contributor=self.user_alice,
                process=self.process,
                size=0,
            )

        data = await create_data()

        # Assert we detect creations.
        self.assertDictEqual(
            json.loads(await client.receive_from()),
            {
                "change_type": ChangeType.CREATE.name,
                "object_id": data.pk,
                "subscription_id": self.subscription_id.hex,
                "source": ["data", data.pk],
            },
        )
        await self.assert_no_more_messages(client)

        # Delete the Data object.
        @database_sync_to_async
        def delete_data(data):
            data.delete()

        data_pk = data.pk
        await delete_data(data)

        # Assert we detect deletions.
        self.assertDictEqual(
            json.loads(await client.receive_from()),
            {
                "change_type": ChangeType.DELETE.name,
                "object_id": data_pk,
                "subscription_id": self.subscription_id.hex,
                "source": ["data", data_pk],
            },
        )
        await self.assert_no_more_messages(client)
        # Assert subscription didn't delete because Data got deleted.
        await self.await_subscription_observer_count(2)

    async def test_observe_entity_table(self):
        """Assert that creation and deletion works on entity table."""
        client = WebsocketCommunicator(self.client_consumer, "/ws/test_session")
        connected, _ = await client.connect()
        self.assertTrue(connected)

        # Create a subscription to the Data and Collection content_type.
        @database_sync_to_async
        def subscribe():
            Subscription.objects.create(
                user=self.user_alice,
                session_id="test_session",
                subscription_id=self.subscription_id,
            ).subscribe(
                content_type=ContentType.objects.get_for_model(Entity),
                object_ids=[Observer.ALL_IDS],
                change_types=[ChangeType.CREATE, ChangeType.DELETE],
            )

        await subscribe()
        await self.await_subscription_observer_count(2)

        # Create a new Entity object.
        @database_sync_to_async
        def create_entity():
            return Entity.objects.create(name="Entity", contributor=self.user_alice)

        # Delete the object.
        @database_sync_to_async
        def delete_object(obj):
            obj.delete()

        @database_sync_to_async
        def set_permission(obj, permission, user):
            obj.set_permission(permission, user)

        @database_sync_to_async
        def modify_superuser(user, issuperuser):
            user.is_superuser = issuperuser
            user.save()

        # First test for regular user.
        # - creation should not sent message
        # - assigning permission must sent create message
        # - removing permission must sent delete message
        # - deleting entity must sent delete message
        entity = await create_entity()

        # No messages when Alice has no permissions.
        await self.assert_no_more_messages(client)

        # CREATE when permission is gained.
        await set_permission(entity, Permission.VIEW, self.user_alice)
        # Assert we detect creations.
        self.assertDictEqual(
            json.loads(await client.receive_from()),
            {
                "change_type": ChangeType.CREATE.name,
                "object_id": entity.pk,
                "subscription_id": self.subscription_id.hex,
                "source": ["entity", entity.pk],
            },
        )
        await self.assert_no_more_messages(client)

        # DELETE when permission is removed.
        await set_permission(entity, Permission.NONE, self.user_alice)
        self.assertDictEqual(
            json.loads(await client.receive_from()),
            {
                "change_type": ChangeType.DELETE.name,
                "object_id": entity.pk,
                "subscription_id": self.subscription_id.hex,
                "source": ["entity", entity.pk],
            },
        )
        await self.assert_no_more_messages(client)

        # CREATE when permission is gained.
        await set_permission(entity, Permission.VIEW, self.user_alice)
        # Assert we detect creations.
        self.assertDictEqual(
            json.loads(await client.receive_from()),
            {
                "change_type": ChangeType.CREATE.name,
                "object_id": entity.pk,
                "subscription_id": self.subscription_id.hex,
                "source": ["entity", entity.pk],
            },
        )
        await self.assert_no_more_messages(client)

        # DELETE when entity is deleted.
        entity_pk = entity.pk
        await delete_object(entity)
        self.assertDictEqual(
            json.loads(await client.receive_from()),
            {
                "change_type": ChangeType.DELETE.name,
                "object_id": entity_pk,
                "subscription_id": self.subscription_id.hex,
                "source": ["entity", entity_pk],
            },
        )
        await self.assert_no_more_messages(client)

        # Repeat for superuser
        # - creation must sent create message
        # - assigning permission must not sent message
        # - removing permission must not sent delete message
        # - deleting entity must sent delete message
        await modify_superuser(self.user_alice, True)

        entity = await create_entity()
        self.assertDictEqual(
            json.loads(await client.receive_from()),
            {
                "change_type": ChangeType.CREATE.name,
                "object_id": entity.pk,
                "subscription_id": self.subscription_id.hex,
                "source": ["entity", entity.pk],
            },
        )
        await self.assert_no_more_messages(client)

        # CREATE when permission is gained.
        await set_permission(entity, Permission.VIEW, self.user_alice)
        await self.assert_no_more_messages(client)

        # No DELETE when permission is removed.
        await set_permission(entity, Permission.NONE, self.user_alice)
        await self.assert_no_more_messages(client)

        # No CREATE when permission is gained.
        await set_permission(entity, Permission.VIEW, self.user_alice)
        # Assert we detect creations.
        await self.assert_no_more_messages(client)

        # DELETE when entity is deleted.
        entity_pk = entity.pk
        await delete_object(entity)
        self.assertDictEqual(
            json.loads(await client.receive_from()),
            {
                "change_type": ChangeType.DELETE.name,
                "object_id": entity_pk,
                "subscription_id": self.subscription_id.hex,
                "source": ["entity", entity_pk],
            },
        )
        await self.assert_no_more_messages(client)

        # Test data creation/deletion in the entity sends message.
        # First for regular user, then superuser.
        await modify_superuser(self.user_alice, False)

        @database_sync_to_async
        def create_data(entity):
            return Data.objects.create(
                name="Test data",
                contributor=self.user_alice,
                process=self.process,
                size=0,
                entity=entity,
            )

        entity = await create_entity()
        await set_permission(entity, Permission.VIEW, self.user_alice)
        self.assertDictEqual(
            json.loads(await client.receive_from()),
            {
                "change_type": ChangeType.CREATE.name,
                "object_id": entity.pk,
                "subscription_id": self.subscription_id.hex,
                "source": ["entity", entity.pk],
            },
        )
        await self.assert_no_more_messages(client)

        data = await create_data(entity)
        self.assertDictEqual(
            json.loads(await client.receive_from()),
            {
                "change_type": ChangeType.CREATE.name,
                "object_id": entity.pk,
                "subscription_id": self.subscription_id.hex,
                "source": ["data", data.pk],
            },
        )
        await self.assert_no_more_messages(client)

        data_pk = data.pk
        await delete_object(data)
        self.assertDictEqual(
            json.loads(await client.receive_from()),
            {
                "change_type": ChangeType.DELETE.name,
                "object_id": entity.pk,
                "subscription_id": self.subscription_id.hex,
                "source": ["data", data_pk],
            },
        )
        await self.assert_no_more_messages(client)

        entity_pk = entity.pk
        await delete_object(entity)
        self.assertDictEqual(
            json.loads(await client.receive_from()),
            {
                "change_type": ChangeType.DELETE.name,
                "object_id": entity_pk,
                "subscription_id": self.subscription_id.hex,
                "source": ["entity", entity_pk],
            },
        )
        await self.assert_no_more_messages(client)

        # Now repeat the test for superuser.
        await modify_superuser(self.user_alice, True)
        entity = await create_entity()
        self.assertDictEqual(
            json.loads(await client.receive_from()),
            {
                "change_type": ChangeType.CREATE.name,
                "object_id": entity.pk,
                "subscription_id": self.subscription_id.hex,
                "source": ["entity", entity.pk],
            },
        )
        await self.assert_no_more_messages(client)

        data = await create_data(entity)
        self.assertDictEqual(
            json.loads(await client.receive_from()),
            {
                "change_type": ChangeType.CREATE.name,
                "object_id": entity.pk,
                "subscription_id": self.subscription_id.hex,
                "source": ["data", data.pk],
            },
        )
        await self.assert_no_more_messages(client)

        data_pk = data.pk
        await delete_object(data)
        self.assertDictEqual(
            json.loads(await client.receive_from()),
            {
                "change_type": ChangeType.DELETE.name,
                "object_id": entity.pk,
                "subscription_id": self.subscription_id.hex,
                "source": ["data", data_pk],
            },
        )
        await self.assert_no_more_messages(client)

        entity_pk = entity.pk
        await delete_object(entity)
        self.assertDictEqual(
            json.loads(await client.receive_from()),
            {
                "change_type": ChangeType.DELETE.name,
                "object_id": entity_pk,
                "subscription_id": self.subscription_id.hex,
                "source": ["entity", entity_pk],
            },
        )
        await self.assert_no_more_messages(client)

    async def test_background_delete(self):
        """Test background delete."""
        client = WebsocketCommunicator(self.client_consumer, "/ws/test_session")
        connected, _ = await client.connect()
        self.assertTrue(connected)

        @database_sync_to_async
        def prepare_test():
            collection = Collection.objects.create(
                name="Collection", contributor=self.user_alice
            )

            entity = Entity.objects.create(
                name="Entity", collection=collection, contributor=self.user_alice
            )

            Data.objects.create(
                name="Data",
                contributor=self.user_alice,
                process=self.process,
                size=0,
                collection=collection,
                entity=entity,
                status=Data.STATUS_DONE,
                output={},
            )
            collection.set_permission(Permission.VIEW, self.user_alice)
            return collection

        collection = await prepare_test()

        # Create a subscription to the Collection content_type.
        @database_sync_to_async
        def subscribe():
            Subscription.objects.create(
                user=self.user_alice,
                session_id="test_session",
                subscription_id=self.subscription_id,
            ).subscribe(
                content_type=ContentType.objects.get_for_model(Collection),
                object_ids=[Observer.ALL_IDS],
                change_types=[ChangeType.CREATE, ChangeType.DELETE],
            )

        await subscribe()
        await self.await_subscription_observer_count(2)

        @database_sync_to_async
        def assert_collection_deleted(collection):
            collection_pk = collection.pk
            entity_pk = collection.entity_set.first().pk
            data_pk = collection.data.first().pk
            collection.delete_background().wait()
            with self.assertRaises(Collection.DoesNotExist):
                collection.refresh_from_db()
            with self.assertRaises(Entity.DoesNotExist):
                Entity.objects.get(pk=entity_pk)
            with self.assertRaises(Data.DoesNotExist):
                Data.objects.get(pk=data_pk)
            return collection_pk, entity_pk, data_pk

        collection_pk, entity_pk, data_pk = await assert_collection_deleted(collection)
        result = [json.loads(await client.receive_from()) for _ in range(3)]
        self.assertCountEqual(
            result,
            [
                {
                    "change_type": ChangeType.DELETE.name,
                    "object_id": collection_pk,
                    "subscription_id": self.subscription_id.hex,
                    "source": ["data", data_pk],
                },
                {
                    "change_type": ChangeType.DELETE.name,
                    "object_id": collection_pk,
                    "subscription_id": self.subscription_id.hex,
                    "source": ["entity", entity_pk],
                },
                {
                    "change_type": ChangeType.DELETE.name,
                    "object_id": collection_pk,
                    "subscription_id": self.subscription_id.hex,
                    "source": ["collection", collection_pk],
                },
            ],
        )
        await self.assert_no_more_messages(client)

    async def test_duplicate(self):
        """Duplicate handles custom signal and has to be test separately."""
        client = WebsocketCommunicator(self.client_consumer, "/ws/test_session")
        connected, _ = await client.connect()
        self.assertTrue(connected)

        # Create a new Collection, Enttity and Data object.
        @database_sync_to_async
        def prepare_test():
            collection = Collection.objects.create(
                name="Collection", contributor=self.user_alice
            )

            entity = Entity.objects.create(
                name="Entity", collection=collection, contributor=self.user_alice
            )

            Data.objects.create(
                name="Data",
                contributor=self.user_alice,
                process=self.process,
                size=0,
                collection=collection,
                entity=entity,
                status=Data.STATUS_DONE,
            )
            collection.set_permission(Permission.VIEW, self.user_alice)
            return collection

        collection = await prepare_test()

        # Create a subscription to the Data and Collection content_type.
        @database_sync_to_async
        def subscribe():
            Subscription.objects.create(
                user=self.user_alice,
                session_id="test_session",
                subscription_id=self.subscription_id,
            ).subscribe(
                content_type=ContentType.objects.get_for_model(Collection),
                object_ids=[Observer.ALL_IDS],
                change_types=[ChangeType.CREATE, ChangeType.DELETE],
            )

        await subscribe()
        await self.await_subscription_observer_count(2)

        @database_sync_to_async
        def duplicate(collection):
            result = collection.duplicate(request_user=self.user_alice).result()
            duplicated_collection = Collection.objects.get(pk__in=result)
            duplicated_entity = duplicated_collection.entity_set.get()
            duplicated_data = duplicated_collection.data.get()
            return duplicated_data, duplicated_entity, duplicated_collection

        duplicated_data, duplicated_entity, duplicated_collection = await duplicate(
            collection
        )
        result = [json.loads(await client.receive_from()) for _ in range(3)]
        self.assertCountEqual(
            result,
            [
                {
                    "change_type": ChangeType.CREATE.name,
                    "object_id": duplicated_collection.pk,
                    "subscription_id": self.subscription_id.hex,
                    "source": ["data", duplicated_data.pk],
                },
                {
                    "change_type": ChangeType.CREATE.name,
                    "object_id": duplicated_collection.pk,
                    "subscription_id": self.subscription_id.hex,
                    "source": ["entity", duplicated_entity.pk],
                },
                {
                    "change_type": ChangeType.CREATE.name,
                    "object_id": duplicated_collection.pk,
                    "subscription_id": self.subscription_id.hex,
                    "source": ["collection", duplicated_collection.pk],
                },
            ],
        )
        await self.assert_no_more_messages(client)

    async def test_observe_collection_table(self):
        """Assert that creation and deletion works on collection table."""
        client = WebsocketCommunicator(self.client_consumer, "/ws/test_session")
        connected, _ = await client.connect()
        self.assertTrue(connected)

        # Create a subscription to the Data and Collection content_type.
        @database_sync_to_async
        def subscribe():
            Subscription.objects.create(
                user=self.user_alice,
                session_id="test_session",
                subscription_id=self.subscription_id,
            ).subscribe(
                content_type=ContentType.objects.get_for_model(Collection),
                object_ids=[Observer.ALL_IDS],
                change_types=[ChangeType.CREATE, ChangeType.DELETE],
            )

        await subscribe()
        await self.await_subscription_observer_count(2)

        # Create a new Collection object.
        @database_sync_to_async
        def create_collection():
            return Collection.objects.create(
                name="Collection", contributor=self.user_alice
            )

        # Delete the object.
        @database_sync_to_async
        def delete_object(obj):
            obj.delete()

        @database_sync_to_async
        def set_permission(obj, permission, user):
            obj.set_permission(permission, user)

        @database_sync_to_async
        def modify_superuser(user, issuperuser):
            user.is_superuser = issuperuser
            user.save()

        @database_sync_to_async
        def create_data(collection):
            return Data.objects.create(
                name="Test data",
                contributor=self.user_alice,
                process=self.process_auto_entity,
                size=0,
                collection=collection,
            )

        # Test that creating data object with auto entity assignes triggers messages.

        # First test for regular user.
        # - collection creation should not sent message
        # - assigning permission on collection must sent create message
        # - creating data object must sent entity and data create messages
        collection = await create_collection()

        # No messages when Alice has no permissions.
        await self.assert_no_more_messages(client)

        # CREATE when permission is gained.
        await set_permission(collection, Permission.VIEW, self.user_alice)
        # Assert we detect creations.
        self.assertDictEqual(
            json.loads(await client.receive_from()),
            {
                "change_type": ChangeType.CREATE.name,
                "object_id": collection.pk,
                "subscription_id": self.subscription_id.hex,
                "source": ["collection", collection.pk],
            },
        )
        await self.assert_no_more_messages(client)

        # We expect 2 messages: one for data and one for the entity.
        data = await create_data(collection)
        entity = data.entity

        self.assertCountEqual(
            [json.loads(await client.receive_from()) for _ in range(2)],
            [
                {
                    "change_type": ChangeType.CREATE.name,
                    "object_id": collection.pk,
                    "subscription_id": self.subscription_id.hex,
                    "source": ["data", data.pk],
                },
                {
                    "change_type": ChangeType.CREATE.name,
                    "object_id": collection.pk,
                    "subscription_id": self.subscription_id.hex,
                    "source": ["entity", entity.pk],
                },
            ],
        )
        await self.assert_no_more_messages(client)

        # Delete the entity object: we expect two messages on collection.
        data_pk = data.pk
        entity_pk = entity.pk
        await delete_object(entity)
        self.assertCountEqual(
            [json.loads(await client.receive_from()) for _ in range(2)],
            [
                {
                    "change_type": ChangeType.DELETE.name,
                    "object_id": collection.pk,
                    "subscription_id": self.subscription_id.hex,
                    "source": ["data", data_pk],
                },
                {
                    "change_type": ChangeType.DELETE.name,
                    "object_id": collection.pk,
                    "subscription_id": self.subscription_id.hex,
                    "source": ["entity", entity_pk],
                },
            ],
        )
        await self.assert_no_more_messages(client)

        # Repeat for superuser.
        await modify_superuser(self.user_alice, True)
        data = await create_data(collection)
        entity = data.entity

        self.assertCountEqual(
            [json.loads(await client.receive_from()) for _ in range(2)],
            [
                {
                    "change_type": ChangeType.CREATE.name,
                    "object_id": collection.pk,
                    "subscription_id": self.subscription_id.hex,
                    "source": ["data", data.pk],
                },
                {
                    "change_type": ChangeType.CREATE.name,
                    "object_id": collection.pk,
                    "subscription_id": self.subscription_id.hex,
                    "source": ["entity", entity.pk],
                },
            ],
        )
        await self.assert_no_more_messages(client)

        # Delete the entity object: we expect two messages on collection.
        data_pk = data.pk
        entity_pk = entity.pk
        await delete_object(entity)
        self.assertCountEqual(
            [json.loads(await client.receive_from()) for _ in range(2)],
            [
                {
                    "change_type": ChangeType.DELETE.name,
                    "object_id": collection.pk,
                    "subscription_id": self.subscription_id.hex,
                    "source": ["data", data_pk],
                },
                {
                    "change_type": ChangeType.DELETE.name,
                    "object_id": collection.pk,
                    "subscription_id": self.subscription_id.hex,
                    "source": ["entity", entity_pk],
                },
            ],
        )
        await self.assert_no_more_messages(client)

    async def test_change_permission_group(self):
        client_bob = WebsocketCommunicator(self.client_consumer, "/ws/test_session_bob")
        connected_bob, _ = await client_bob.connect()
        self.assertTrue(connected_bob)

        client_alice = WebsocketCommunicator(
            self.client_consumer, "/ws/test_session_alice"
        )
        connected_alice, _ = await client_alice.connect()
        self.assertTrue(connected_alice)

        # Create a Data object visible to Bob.
        @database_sync_to_async
        def create_data():
            self.collection = Collection.objects.create(
                contributor=self.user_alice,
                name="Test collection",
            )
            self.collection.set_permission(Permission.VIEW, self.user_bob)
            self.collection.set_permission(Permission.VIEW, self.user_alice)

            self.collection2 = Collection.objects.create(
                contributor=self.user_alice,
                name="Test collection 2",
            )
            self.collection2.set_permission(Permission.NONE, self.user_bob)
            self.collection2.set_permission(Permission.VIEW, self.user_alice)

            data = Data.objects.create(
                pk=42,
                name="Test data",
                slug="test-data",
                contributor=self.user_alice,
                process=self.process,
                collection=self.collection,
                size=0,
            )
            return data

        data = await create_data()

        # Create a subscription to the Data object by Bob.
        @database_sync_to_async
        def subscribe(data):
            subscription_bob = Subscription.objects.create(
                user=self.user_bob,
                session_id="test_session_bob",
                subscription_id=self.subscription_id,
            )
            subscription_bob.subscribe(
                content_type=ContentType.objects.get_for_model(Data),
                object_ids=[42],
                change_types=[ChangeType.UPDATE, ChangeType.DELETE],
            )
            subscription_bob.subscribe(
                content_type=ContentType.objects.get_for_model(Collection),
                object_ids=[self.collection.pk, self.collection2.pk],
                change_types=[ChangeType.UPDATE, ChangeType.DELETE, ChangeType.CREATE],
            )

            subscription_alice = Subscription.objects.create(
                user=self.user_alice,
                session_id="test_session_alice",
                subscription_id=self.subscription_id2,
            )
            subscription_alice.subscribe(
                content_type=ContentType.objects.get_for_model(Collection),
                object_ids=[self.collection.pk, self.collection2.pk],
                change_types=[ChangeType.DELETE, ChangeType.CREATE, ChangeType.UPDATE],
            )
            subscription_alice.subscribe(
                content_type=ContentType.objects.get_for_model(Data),
                object_ids=[data.pk],
                change_types=[ChangeType.DELETE, ChangeType.CREATE, ChangeType.UPDATE],
            )

        await subscribe(data)

        # Reset the PermissionGroup of the Data object (removes permissions to Bob)
        @database_sync_to_async
        def change_permission_group(data):
            data.move_to_collection(self.collection2)

        await change_permission_group(data)
        notifications_bob = [
            json.loads(await client_bob.receive_from()) for _ in range(2)
        ]
        notifications_alice = [
            json.loads(await client_alice.receive_from()) for _ in range(3)
        ]

        # Assert that Bob sees this as a deletion.
        self.assertCountEqual(
            notifications_bob,
            [
                {
                    "change_type": ChangeType.DELETE.name,
                    "object_id": 42,
                    "subscription_id": self.subscription_id.hex,
                    "source": ["data", 42],
                },
                {
                    "object_id": self.collection.pk,
                    "change_type": ChangeType.DELETE.name,
                    "source": ["data", 42],
                    "subscription_id": self.subscription_id.hex,
                },
            ],
        )
        self.assertCountEqual(
            notifications_alice,
            [
                {
                    "object_id": data.pk,
                    "change_type": ChangeType.UPDATE.name,
                    "subscription_id": self.subscription_id2.hex,
                    "source": ["data", 42],
                },
                {
                    "object_id": self.collection.pk,
                    "change_type": ChangeType.DELETE.name,
                    "source": ["data", 42],
                    "subscription_id": self.subscription_id2.hex,
                },
                {
                    "object_id": self.collection2.pk,
                    "change_type": ChangeType.CREATE.name,
                    "source": ["data", 42],
                    "subscription_id": self.subscription_id2.hex,
                },
            ],
        )
        await self.assert_no_more_messages(client_bob)
        await self.assert_no_more_messages(client_alice)

    async def test_modify_permissions(self):
        client = WebsocketCommunicator(self.client_consumer, "/ws/test_session")
        connected, details = await client.connect()
        self.assertTrue(connected)

        # Create a new Data object.
        @database_sync_to_async
        def create_data():
            return Data.objects.create(
                pk=42,
                name="Test data",
                slug="test-data",
                contributor=self.user_alice,
                process=self.process,
                size=0,
            )

        data = await create_data()

        # Create a subscription to the Data content_type by Bob.
        @database_sync_to_async
        def subscribe():
            Subscription.objects.create(
                user=self.user_bob,
                session_id="test_session",
                subscription_id=self.subscription_id,
            ).subscribe(
                content_type=ContentType.objects.get_for_model(Data),
                object_ids=[Observer.ALL_IDS],
                change_types=[ChangeType.CREATE, ChangeType.UPDATE, ChangeType.DELETE],
            )

        await subscribe()
        await self.await_subscription_observer_count(3)

        # Grant Bob view permissions to the Data object.
        @database_sync_to_async
        def grant_permissions(data):
            data.set_permission(Permission.VIEW, self.user_bob)

        await grant_permissions(data)

        # Assert we detect gaining permissions as creations.
        self.assertDictEqual(
            json.loads(await client.receive_from()),
            {
                "change_type": ChangeType.CREATE.name,
                "object_id": 42,
                "subscription_id": self.subscription_id.hex,
                "source": ["data", 42],
            },
        )
        await self.assert_no_more_messages(client)

        # Revoke permissions for Bob.
        @database_sync_to_async
        def revoke_permissions(data):
            data.set_permission(Permission.NONE, self.user_bob)

        await revoke_permissions(data)

        # Assert we detect losing permissions as deletions.
        self.assertDictEqual(
            json.loads(await client.receive_from()),
            {
                "change_type": ChangeType.DELETE.name,
                "object_id": 42,
                "subscription_id": self.subscription_id.hex,
                "source": ["data", 42],
            },
        )
        await self.assert_no_more_messages(client)

    async def test_observe_table_no_permissions(self):
        client = WebsocketCommunicator(self.client_consumer, "/ws/test_session")
        connected, details = await client.connect()
        self.assertTrue(connected)

        # Create a subscription to the Data content_type by Bob.
        @database_sync_to_async
        def subscribe():
            Subscription.objects.create(
                user=self.user_bob,
                session_id="test_session",
                subscription_id=self.subscription_id,
            ).subscribe(
                content_type=ContentType.objects.get_for_model(Data),
                object_ids=[Observer.ALL_IDS],
                change_types=[ChangeType.CREATE, ChangeType.UPDATE, ChangeType.DELETE],
            )

        await subscribe()
        await self.await_subscription_observer_count(3)

        # Create a new Data object.
        @database_sync_to_async
        def create_data():
            data = Data.objects.create(
                pk=42,
                name="Test data",
                slug="test-data",
                contributor=self.user_alice,
                process=self.process,
                size=0,
            )
            return data

        data = await create_data()

        # Assert we don't detect creations (Bob doesn't have permissions).
        await self.assert_no_more_messages(client)

        # Delete the Data object.
        @database_sync_to_async
        def delete_data(data):
            data.delete()

        await delete_data(data)

        # Assert we don't detect deletions.
        await self.assert_no_more_messages(client)

        # Assert subscription didn't delete.
        await self.await_subscription_observer_count(3)


class BackgroundTaskTestCase(TransactionResolweAPITestCase):
    def setUp(self):
        self.viewset = BackgroundTaksViewSet
        self.resource_name = "backgroundtask"
        super().setUp()
        self.task_preparing = BackgroundTask.objects.create(
            status=BackgroundTask.STATUS_WAITING,
            description="Preparing",
            output="Output preparing",
        )
        self.task_running = BackgroundTask.objects.create(
            status=BackgroundTask.STATUS_PROCESSING,
            description="Processing",
            output="Output processing",
        )
        self.task_ok = BackgroundTask.objects.create(
            status=BackgroundTask.STATUS_DONE, description="Done", output="Output done"
        )
        self.task_error = BackgroundTask.objects.create(
            status=BackgroundTask.STATUS_ERROR,
            description="Error",
            output="Output error",
        )

        User = get_user_model()
        self.alice = User.objects.create(
            username="alice",
            email="alice@test.com",
            first_name="Ana",
            last_name="Banana",
        )
        # Alice has access to all the tasks by default.
        self.task_preparing.set_permission(Permission.VIEW, self.alice)
        self.task_running.set_permission(Permission.VIEW, self.alice)
        self.task_ok.set_permission(Permission.VIEW, self.alice)
        self.task_error.set_permission(Permission.VIEW, self.alice)

    def test_background_task_list(self):
        # Public access should be empty.
        result = self._get_list()
        self.assertEqual(result.status_code, status.HTTP_200_OK)
        self.assertEqual(result.data, [])

        result = BackgroundTaskSerializer(self.task_preparing).data

        # Alice should see the preparing and running tasks by default.
        expected = [
            BackgroundTaskSerializer(self.task_preparing).data,
            BackgroundTaskSerializer(self.task_running).data,
        ]
        result = self._get_list(
            user=self.alice,
            query_params={
                "status__in": f"{BackgroundTask.STATUS_WAITING},{BackgroundTask.STATUS_PROCESSING}"
            },
        )
        self.assertEqual(result.status_code, status.HTTP_200_OK)
        self.assertCountEqual(result.data, expected)

        # Filter task by the specific id.
        expected = [BackgroundTaskSerializer(self.task_preparing).data]
        result = self._get_list(
            user=self.alice,
            query_params={"id": self.task_preparing.id},
        )
        self.assertEqual(result.status_code, status.HTTP_200_OK)
        self.assertCountEqual(result.data, expected)

        # Filter task by the list of ids.
        expected = [
            BackgroundTaskSerializer(self.task_preparing).data,
            BackgroundTaskSerializer(self.task_running).data,
        ]
        result = self._get_list(
            user=self.alice,
            query_params={"id__in": f"{self.task_preparing.id},{self.task_running.id}"},
        )
        self.assertEqual(result.status_code, status.HTTP_200_OK)
        self.assertCountEqual(result.data, expected)

        # Filter only tasks in status OK and ERROR.
        result = self._get_list(
            user=self.alice,
            query_params={
                "status__in": f"{BackgroundTask.STATUS_DONE},{BackgroundTask.STATUS_ERROR}",
            },
        )
        expected = [
            BackgroundTaskSerializer(self.task_ok).data,
            BackgroundTaskSerializer(self.task_error).data,
        ]
        self.assertEqual(result.status_code, status.HTTP_200_OK)
        self.assertCountEqual(result.data, expected)

        # Alice should not see tasks without explicit permissions.
        self.task_running.set_permission(Permission.NONE, self.alice)
        expected = [BackgroundTaskSerializer(self.task_preparing).data]
        result = self._get_list(
            user=self.alice,
            query_params={
                "status__in": f"{BackgroundTask.STATUS_WAITING},{BackgroundTask.STATUS_PROCESSING}"
            },
        )
        self.assertEqual(result.status_code, status.HTTP_200_OK)
        self.assertCountEqual(result.data, expected)

        # Alice should see all tasks if superuser.
        self.alice.is_superuser = True
        self.alice.is_staff = True
        self.alice.save()
        expected = [
            BackgroundTaskSerializer(self.task_preparing).data,
            BackgroundTaskSerializer(self.task_running).data,
            BackgroundTaskSerializer(self.task_ok).data,
            BackgroundTaskSerializer(self.task_error).data,
        ]
        result = self._get_list(user=self.alice)
        self.assertEqual(result.status_code, status.HTTP_200_OK)
        self.assertCountEqual(result.data, expected)


class ObserverAPITestCase(TransactionResolweAPITestCase):
    def setUp(self):
        self.viewset = DataViewSet
        self.resource_name = "data"
        super().setUp()
        self.list_view = self.viewset.as_view({"post": "subscribe"})

        user_model = get_user_model()
        self.user_alice = user_model.objects.create(
            username="alice",
            email="alice@test.com",
            first_name="Ana",
            last_name="Banana",
        )
        self.user_bob = user_model.objects.create(
            username="capital-bob",
            email="bob@bob.bob",
            first_name="Capital",
            last_name="Bobnik",
        )
        self.process = Process.objects.create(
            name="Dummy process", contributor=self.user_alice
        )
        Data.objects.create(
            pk=42,
            name="Test data",
            slug="test-data",
            contributor=self.user_alice,
            process=self.process,
            size=0,
        )
        self.client_consumer = URLRouter(
            [path("ws/<slug:subscriber_id>", ClientConsumer().as_asgi())]
        )

    def test_subscribe(self):
        # Subscribe to model updates.
        resp = self._post(user=self.user_alice, data={"session_id": "test"})
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertDictEqual(
            resp.data,
            {"subscription_id": Subscription.objects.all()[0].subscription_id.hex},
        )
        # There should be 2 subscriptions: for CREATE and DLETE.
        self.assertEqual(Observer.objects.count(), 2)
        self.assertEqual(
            Observer.objects.filter(
                change_type=ChangeType.CREATE.value,
                object_id=Observer.ALL_IDS,
                content_type=ContentType.objects.get_for_model(Data),
            ).count(),
            1,
        )

        # Subscribe to instance updates.
        resp = self._post(
            user=self.user_alice, data={"session_id": "test", "ids": [42]}
        )
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        sub_qs = Subscription.objects.filter(
            session_id="test", observers__object_id=42
        ).distinct()
        self.assertEqual(sub_qs.count(), 1)
        self.assertDictEqual(
            resp.data,
            {"subscription_id": sub_qs.first().subscription_id.hex},
        )
        self.assertEqual(Observer.objects.count(), 5)
        self.assertEqual(
            Observer.objects.filter(
                change_type=ChangeType.UPDATE.value,
                object_id=42,
                content_type=ContentType.objects.get_for_model(Data),
            ).count(),
            1,
        )
        self.assertEqual(
            Observer.objects.filter(
                change_type=ChangeType.DELETE.value,
                object_id=42,
                content_type=ContentType.objects.get_for_model(Data),
            ).count(),
            1,
        )

        # Re-subscribe to the same endpoint.
        resp = self._post(
            user=self.user_alice, data={"session_id": "test", "ids": [42]}
        )
        # Assert we don't have duplicate observers.
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(Observer.objects.count(), 5)

    def test_subscribe_to_forbidden_object(self):
        resp = self._post(user=self.user_bob, data={"session_id": "test", "ids": [42]})
        self.assertEqual(resp.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(Observer.objects.count(), 0)

    def test_subscribe_to_nonexistent_object(self):
        resp = self._post(
            user=self.user_alice, data={"session_id": "test", "ids": [111]}
        )
        self.assertEqual(resp.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(Observer.objects.count(), 0)

    def test_subscribe_anonymously(self):
        # A different anonymous user was created with each test, so the
        # ANONYMOUS_USER global variable is cached incorrectly and must be reset.
        anon = get_anonymous_user(cache=False)

        # Make data public.
        data = Data.objects.get(pk=42)
        data.set_permission(Permission.VIEW, anon)

        resp = self._post(data={"session_id": "test", "ids": [42]})
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertDictEqual(
            resp.data,
            {"subscription_id": Subscription.objects.all()[0].subscription_id.hex},
        )
        self.assertEqual(
            Observer.objects.filter(
                change_type=ChangeType.UPDATE.value,
                object_id=42,
                content_type=ContentType.objects.get_for_model(Data),
            ).count(),
            1,
        )
        self.assertEqual(
            Observer.objects.filter(
                change_type=ChangeType.DELETE.value,
                object_id=42,
                content_type=ContentType.objects.get_for_model(Data),
            ).count(),
            1,
        )



from django.urls import include, path
from drf_spectacular.views import SpectacularAPIView, SpectacularSwaggerView


urlpatterns = [
    path("", include("resolwe.api_urls")),
    path("api-auth/", include("rest_framework.urls", namespace="rest_framework")),
    path("api/schema/", SpectacularAPIView.as_view(), name="schema"),
    path(
        "api/schema/swagger-ui/",
        SpectacularSwaggerView.as_view(url_name="schema"),
        name="swagger-ui",
    ),
]
